package physbeans.editors;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.beans.*;
import physbeans.inout.*;
import physbeans.math.*;
import physbeans.model.*;
import static java.lang.Math.*;

/**
 * The panel inside the ArrayOfColorEditor. It contains
 * a list of the array values as triples of r/g/b values.
 */
public class ArrayOfColorCustomEditor extends JPanel {
  
  protected PropertyEditorSupport editor;
  protected NumericTextField[] textfieldR, textfieldG, textfieldB;
  protected NumericTextField dimField;
  protected Box pList;
  protected Font standardFont;
  
  protected final int cols = 6;   // size of text fields
  protected final int prec = 3;   // precision of text fields
  
  public ArrayOfColorCustomEditor(PropertyEditorSupport ed) {
    // get the old values
    editor = ed;
    Color[] value = (Color[]) editor.getValue();
    int dim = value.length;
    
    // vertical list with values, on top box for dimension
    standardFont = new Font("Lucida Sans", Font.PLAIN, 11);
    setLayout(new BorderLayout());
    JPanel pHead = new JPanel();
    dimField = new NumericTextField(dim, 8, 4, "Dimension = ", "");
    dimField.setFont(standardFont);
    pHead.add(dimField);
    dimField.addPropertyChangeListener(new MyDimListener());
    
    // create panel with text fields
    // layout: horizontal boxes with x-y pairs, stacked in a vertical box
    pList = new Box(BoxLayout.Y_AXIS);
    pList.setPreferredSize(new Dimension(200, dim*20+20));
    JScrollPane scrollPane = new JScrollPane(pList,
        JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
        JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    textfieldR = new NumericTextField[0];
    setTextFieldsToVector(value);
    
    add(pHead, BorderLayout.NORTH);
    add(scrollPane, BorderLayout.CENTER);
    setMaximumSize(new Dimension(200, 150));
  }
  
  /**
   * Called when the value of a textfield is changed.
   */
  protected void changeValue() {
    Color[] newValue = getValueFromTextFields();
    
    editor.setValue(newValue);
    editor.firePropertyChange();
  }
  
  /**
   * Called when the value of the dimension field is changed.
   */
  protected void resize() {
    int newDim = (int) Math.round(dimField.getValue());
    
    if (newDim < 0) {
      newDim = textfieldR.length;
    }
    
    if (newDim != textfieldR.length) {
      Color[] newValue = new Color[newDim];
      for (int i = 0; i < newDim; i++) {
        newValue[i] = new Color(255, 255, 255);
      }
      setTextFieldsToVector(newValue);
      editor.setValue(newValue);
      editor.firePropertyChange();
    }
    
    dimField.setValueInternally(newDim);   // to fix fractional input
  }
  
  /**
   * Gets Color array from the text fields.
   */
  protected Color[] getValueFromTextFields() {
    int dim = textfieldR.length;
    Color[] colors = new Color[dim];
    
    for (int i = 0; i < dim; i++) {
      int r = (int) max(min(textfieldR[i].getValue(), 255), 0);
      int g = (int) max(min(textfieldG[i].getValue(), 255), 0);
      int b = (int) max(min(textfieldB[i].getValue(), 255), 0);
      colors[i] = new Color(r, g, b);
    }
    
    return colors;
  }
  
  /**
   * Sets text fields from Color array.
   */
  protected void setTextFieldsToVector(Color[] newVal) {
    int dim = newVal.length;
    
    if (textfieldR.length == dim) {
      for (int i = 0; i < dim; i++) {
        Color c = newVal[i];
        textfieldR[i].setValueInternally(c.getRed());
        textfieldG[i].setValueInternally(c.getGreen());
        textfieldB[i].setValueInternally(c.getBlue());
      }
    } else {
      pList.removeAll();
      Box[] xBox = new Box[dim];
      textfieldR = new NumericTextField[dim];
      textfieldG = new NumericTextField[dim];
      textfieldB = new NumericTextField[dim];
      
      for (int i = 0; i < dim; i++) {
        Color c = newVal[i];
        xBox[i] = new Box(BoxLayout.X_AXIS);
        textfieldR[i] = new NumericTextField(c.getRed(), cols, prec, "R = ", "");
        textfieldG[i] = new NumericTextField(c.getGreen(), cols, prec, "G = ", "");
        textfieldB[i] = new NumericTextField(c.getBlue(), cols, prec, "B = ", "");
        textfieldR[i].setFont(standardFont);
        textfieldG[i].setFont(standardFont);
        textfieldB[i].setFont(standardFont);
        
        xBox[i].add(textfieldR[i]);
        xBox[i].add(textfieldG[i]);
        xBox[i].add(textfieldB[i]);
        pList.add(xBox[i]);
        textfieldR[i].addPropertyChangeListener(new MyTextListener());
        textfieldG[i].addPropertyChangeListener(new MyTextListener());
        textfieldB[i].addPropertyChangeListener(new MyTextListener());
      }
      pList.setPreferredSize(new Dimension(200, dim*20+20));
    }
    validate();
  }
  
  
  /**
   * Listeners to change in the text fields.
   */
  public class MyTextListener implements PropertyChangeListener {
    public void propertyChange(PropertyChangeEvent e) {
      changeValue();
    }
  }
  
  /**
   * Listeners to change in the dimension field.
   */
  public class MyDimListener implements PropertyChangeListener {
    public void propertyChange(PropertyChangeEvent e) {
      resize();
    }
  }
}
