package physbeans.func;

import physbeans.math.*;

/**
 * Combines two input vectors of equal length into one output vector of the same
 * length by choosing the components from one or the other input vector according
 * to an internal boolean array.
 * <p>The boolean array usually has the same length as the input vectors. If it
 * is larger, the extra entries are ignored, if it is smaller, its entries are
 * used cyclically. In the special case of length 1 it behaves like a
 * VectorSwitch with two input vectors that is switched by a boolean instead
 * of an integer index.
 */
public class MergeFunction extends GenericVectorFunction {
  
  protected boolean[] useFirstVector;
  
  /**
   * Creates MergeFunction with variable input length, all components come from
   * the first vector.
   */
  public MergeFunction()  {
    super(2, 1, -1);
    setCombineMode(true);
    useFirstVector = new boolean[]{true};
  }
  
  /**
   * Computes new values for output vector.
   */
  protected void compute() {
    if (outputValues == null || outputValues[0] == null) {
      return;
    }
    
    int n = outputValues[0].getDimension();
    
    for (int i = 0; i < n; i++) {
      if (useFirstVector[i % useFirstVector.length]) {
        outputValues[0].set(i, inputValues[0].get(i));
      } else {
        outputValues[0].set(i, inputValues[1].get(i));
      }
    }
  }
  
  /**
   * Gets boolean value that switches the vector component of given index.<br>
   * If true, the corresponding component comes from the first vector, else
   * from the second.
   */
  public boolean getUseFirstVector(int index) {
    return useFirstVector[index];
  }
  
  /**
   * Gets boolean array that switches the vector components.<br>
   * If true, the corresponding component comes from the first vector, else
   * from the second.
   */
  public boolean[] getUseFirstVector() {
    return useFirstVector;
  }
  
  /**
   * Sets boolean value that switches the vector component of given index.<br>
   * If true, the corresponding component comes from the first vector, else
   * from the second.
   */
  public void setUseFirstVector(int index, boolean useFirstVector) {
    this.useFirstVector[index] = useFirstVector;
    inform();
  }
  
  /**
   * Sets processed flags of all inputs.<br>
   * If true, the corresponding input signal is processed, otherwise simply copied.
   * <p>If the array is shorter than the output vector, it is used cyclically.
   */
  public void setUseFirstVector(boolean[] useFirstVector) {
    this.useFirstVector = useFirstVector;
    inform();
  }
  
  /**
   * Sets switches for the vector component of given index using a double value<br>
   * If val > 0, the corresponding component comes from the first vector, else
   * from the second.
   * <p>Interface function to make doubles work as booleans.
   */
  public void setUseFirstVectorFromDouble(int index, double val) {
    boolean b = (val > 0)? true : false;
    setUseFirstVector(index, b);
  }
  
  /**
   * Sets processed flags of all inputs using a DVector instead a boolean array.
   */
  public void setUseFirstVectorFromDVector(DVector useFirstVector) {
    setUseFirstVector(useFirstVector.toBoolean());
  }
  
  /**
   * Sets value of first input vector.
   * <P>Conveniance function to simplify routing.
   */
  public void setFirstVector(DVector v) {
    setInputVector(0, v);
  }
  
  /**
   * Sets value of second input vector.
   * <P>Conveniance function to simplify routing.
   */
  public void setSecondVector(DVector v) {
    setInputVector(1, v);
  }
}
