package physbeans.func;

import java.io.*;
import physbeans.event.*;
import physbeans.math.*;

/**
 * Merges several scalar inputs into one output vector.<br>
 * Forwards the new vector each time all inputs have got new values.<P>
 * If all inputs are set with one vector, it fires immediately.<br>
 * When trigger() is called explicitly it fires regardless of the input state.
 */
public class Multiplexer extends GenericNdFunction implements Serializable {

  protected StateCombiner states;
  protected boolean collectInput;

  /** Creates a new instance of MergeVectorFunction */
  public Multiplexer() {
    this(2);  // merge two scalars
  }

  /**
   *  Creates a Multiplexer with n scalar inputs.
   */
  public Multiplexer(int nIn) {
    // uses the generic input vector for the scalar values
    super(nIn, nIn);
    states = new StateCombiner(nIn);
    collectInput = true;
  }

  /**
   * computes new values for output vector.
   */
  protected DVector computeFunction(DVector in) {
    // simply propagates values to the output vector
    return in.copy();
  }

  /**
   * propagates input values
   */
  public void trigger() {
    compute();  // compute new output values
    fireVectorEvent(new VectorEvent(this, outputValues)); // inform listeners
  }

  /**
   * Sets the number of input values.<br>
   */
  public void setInputLength(int size) {
    inputValues = new DVector(size);
    setOutputLength(size);
    states = new StateCombiner(size);
  }

  /**
   * Sets new input value.<br>
   * If autotrigger is true (default), it triggers after new values have arrived
   * for each input parameter.
   * If one sets collectInput to false, it trigger after every new value.
   */
  public void setInputValues(int index, double val) {
    inputValues.set(index, val);
    states.setState(index);

    if (autoTriggered && (!collectInput || states.getCombinedState())) {
      trigger();
      if (collectInput) {
        states.reset();
      }
    }
  }

  /**
   * Sets new input values from a vector.<br>
   * If n = length(v), it sets values index, index+1, .., index+n-1 .<br>
   * If autotrigger is true (default), it triggers after new values have arrived
   * for each input parameter.
   * If one sets collectInput to false, it trigger after every new value.
   */
  public void setInputValueFromVector(int index, DVector v) {
    for (int i = 0; i < v.getDimension(); i++) {
      inputValues.set(index + i, v.get(i));
      states.setState(index + i);
    }

    if (autoTriggered && (!collectInput || states.getCombinedState())) {
      trigger();
      if (collectInput) {
        states.reset();
      }
    }
  }

  /**
   * Gets the combine mode flag.
   * <p>If it is true (as well as autoTriggered) then the function triggers only
   * after all input vectors have been set. Otherwise each setting of an input
   * vector leads to a trigger event.
   * <p>After a trigger new values for all inputs have to arrive for the next
   * trigger.
   */
  public boolean isCollectInput() {
    return collectInput;
  }

  /**
   * Sets the combine mode flag.
   * <p>If it is true (as well as autoTriggered) then the function triggers only
   * after all input vectors have been set. Otherwise each setting of an input
   * vector leads to a trigger event.
   * <p>After a trigger new values for all inputs have to arrive for the next
   * trigger.
   */
  public void setCollectInput(boolean b) {
    collectInput = b;
  }
}
