package physbeans.func;

import java.io.*;
import physbeans.math.*;

/**
 * Generic function that is defined by a parsed string argument.
 *
 * <p>One first defines a set of variable names (f.i. "x", "y", "A"), then the
 * function string (like "x + A*sin(y)", "sqrt(x) + A^y"). This creates a
 * function with a threedimensional input vector and a twodimensional output 
 * vector.
 *
 * <p>Additionally one can set the last few variable names as parameters by 
 * defining parameter values, f.i.
 * <pre>
 *     variableName: {"t", "A", "omega"}
 *     function:  {"A*cos(omega*t)", "A*sin(omega*t)"}
 *     parameter: {2.0, 0.4}</pre>
 * This defines a function of the remaining variables, here just t, with a 1d
 * input vector and a 2d output vector. The parameters A and omega are set to the
 * given values and can be changed using the setParameter methods.
 * <p>
 * <p>Known constants: 
 * <pre>
 *    pi, e, infinity
 * </pre>
 *
 * <p>Known functions of one variable:
 * <pre>
 *    sin, cos, tan, log, log10, abs, int, frac, asin, acos, atan, sinh, cosh, 
 *    tanh, asinh, acosh, atanh, ceil, floor, round, exp, sqr, sqrt, 
 *    sign, step, random
 * </pre>
 * <p>Known functions of two variables:
 * <pre>
 *    min, max, mod, atan2
 * </pre>
 *
 * <p>Special function if:
 * <pre>
 *    if(condition, expression1, expression2)</pre>
 * returns expression1, if condition is true, expression2 otherwise.
 *
 * <p>Function strings can contain abbreviations for long expressions in a
 * backwards manner, f.i.
 * <pre>
 *    a/b; a:sin(c); b:cos(c); c:x*x</pre> 
 * defines the function tan(x*x).
 */
public class ParsedFunction extends GenericNdFunction implements Serializable {
  
  protected String[] function;
  protected String[] vars;
  protected DVector parameter;
  
  protected SuryonoParser[] parser;
  
  /** Creates a new instance of ParsedFunction */
  public ParsedFunction() {
    super(1, 1);
    
    vars = new String[] {"x"};
    function = new String[] {"tan(x)"};
    parameter = new DVector(0);
    
    createParsers();
  }
  
  
  protected DVector computeFunction(DVector in) {
    int nVars = vars.length;
    int nFunc = function.length;
    int nPars = parameter.getDimension();
    DVector result = new DVector(nFunc);
    
    for (int f = 0; f < nFunc; f++) {
      // set variables from input vector
      for (int v = 0; v < nVars - nPars; v++) {
        parser[f].setVariable(v+1, in.get(v));
      }
      
      // set parameters
      for (int p = 0; p < nPars; p++) {
        parser[f].setVariable(nVars - nPars + p + 1, parameter.get(p));
      }
      
      result.set(f, parser[f].evaluate());
    }
    
    return result;
  }
  
  /**
   * Gets description string of function with given index.
   */
  public String getFunction(int i) {
    return function[i];
  }
  
  /**
   * Gets description strings of all functions.
   */
  public String[] getFunction() {
    return function;
  }
  
  /**
   * Sets description string of function with given index.
   */
  public void setFunction(int i, String function) {
    this.function[i] = function;
    parser[i].define(function);
    parser[i].parse();
  }
  
  /**
   * Sets description strings of all functions.
   */
  public void setFunction(String[] function) {
    this.function = function;
    createParsers();
    setOutputLength(function.length);
  }
  
  /**
   * Gets the name of variable i.
   */
  public String getVariableName(int i) {
    return vars[i];
  }
  
  /**
   * Gets the names of all variables.
   */
  public String[] getVariableName() {
    return vars;
  }
  
  /**
   * Sets the name of variable i.
   */
  public void setVariableName(int i, String name) {
    vars[i] = name;
    createParsers();
  }
  
  /**
   * Sets the names of all variables.
   */
  public void setVariableName(String[] names) {
    vars = names;
    createParsers();
    setInputLength(vars.length - parameter.getDimension());
  }
  
  
  /**
   * Gets the value of parameter i.
   */
  public double getParameter(int i) {
    return parameter.get(i);
  }
  
  /**
   * Gets the values of all parameters.
   */
  public DVector getParameter() {
    return parameter;
  }
  
  /**
   * Sets the value of parameter i.
   */
  public void setParameter(int i, double val) {
    parameter.set(i, val);
  }
  
  /**
   * Sets the values of all parameters.<br>
   * The number of parameters must be larger than the number of variables.
   */
  public void setParameter(DVector vals) {
    // problem if parameters are set before variables
    // TODO good solution: custom editor for ParsedFunction
    parameter = vals;
    int nPars = vals.getDimension();
    if (nPars <= vars.length) {
      setInputLength(vars.length - nPars);
    }
  }
  
  /**
   * Creates parsers for the current number of variables and functions.
   */
  protected void createParsers() {
    int nFunc = function.length;
    int nVars = vars.length;
    
    if (parser == null || nFunc != parser.length) {
      parser = new SuryonoParser[nFunc];
      for (int i = 0; i < nFunc; i++) {
        parser[i] = new SuryonoParser(nVars);
      }
    } else if (nVars != parser[0].getVariableNames().length) {
      for (int i = 0; i < nFunc; i++) {
        parser[i] = new SuryonoParser(nVars);
      }
    }
    
    for (int f = 0; f < nFunc; f++) {
      for (int v = 0; v < nVars; v++) {
        parser[f].defineVariable(v+1,vars[v]);
      }
      parser[f].define(function[f]);
      parser[f].parse();
    }
  }
  
}
