package physbeans.func;

import physbeans.math.*;

/**
 * Function of several input vectors v1, v2, .. and several output vectors (all of
 * the same length) , which is defined by parsing strings containing the
 * symbols v1, v2 ... as variable names.
 * <p>The function works pointwise, i.e. element i of the output vectors
 * is computed by applying the parsed functions to the input elements v1(i),
 * v2(i)...
 * <p>
 * For details about the parser
 * @see physbeans.func.ParsedFunction
 */
public class ParsedVectorFunction extends GenericVectorFunction {
  
  protected ParsedFunction parseFunc;
  
  /** Creates a new instance of ParsedVectorFunction */
  public ParsedVectorFunction() {
    super(2, 1, 100);
    
    parseFunc = new ParsedFunction();
    defineVariables();
    setFunction(new String[] {"v1 + v2"});
  }
  
  protected void compute() {
    int len = inputValues[0].getDimension();
    int nIn = inputValues.length;
    int nOut = outputValues.length;
    DVector inVec = new DVector(nIn);
    DVector outVec = new DVector(nOut);
    
    for (int i = 0; i < len; i++) {
      // assemble input values
      for (int k = 0; k < nIn; k++) {
        inVec.set(k, inputValues[k].get(i));
      }
      
      parseFunc.setInputVector(inVec);
      outVec = parseFunc.getOutputVector();
      
      // distribute output values
      for (int k = 0; k < nOut; k++) {
        outputValues[k].set(i, outVec.get(k));
      }      
    }
  }
  
  /**
   * Sets the number of input vectors.
   */
  public void setNInput(int nIn) {
    inputValues = new DVector[nIn];
    defineVariables();
  }
  
  /**
   * Gets description strings of all functions.
   * <p>Each string defines one element of the output vector.
   */
  public String[] getFunction() {
    return parseFunc.getFunction();
  }
  
  /**
   * Sets description strings of all functions.
   * <p>Each string defines one element of the output vector.
   */
  public void setFunction(String[] function) {
    parseFunc.setFunction(function) ;
    super.setNOutput(function.length);
    for (int i = 0; i < function.length; i++) {
      setOutputLength(i, -1);
    }
  }
  
  /**
   * Defines the variable names v1, v2, ...vn, that are used for parsing the
   * strings.
   */
  protected void defineVariables() {
    int nVars = getNInput();
    String[] vars = new String[nVars];
    
    for (int i = 1; i <= nVars; i++) {
      vars[i-1] = "v" + i;
    }
    
    parseFunc.setVariableName(vars);
  }
}
