package physbeans.func;

import physbeans.math.*;

/**
 * Scales input values from a given input interval to an output interval.
 * Scaling can be linear or logarithmic
 */
public class ScalingFunction extends OneParameterFunction {
  
  protected double xMin, xMax;
  protected double yMin, yMax;
  protected boolean logarithmic, saturated;
  
  /** Creates a new instance of ScalingFunction */
  public ScalingFunction() {
    super(-1);
    
    xMin = -1.0;
    xMax =  1.0;
    yMin =  0.0;
    yMax = 10.0;
    logarithmic = false;
    saturated = true;
  }
  
  /**
   * Gets minimal input value.
   */
  public double getXMin() {
    return this.xMin;
  }
  
  /**
   * Sets minimal input value.
   */
  public void setXMin(double xMin) {
    this.xMin = xMin;
  }
  
  /**
   * Gets maximal input value.
   */
  public double getXMax() {
    return this.xMax;
  }
  
  /**
   * Sets maximal input value.
   */
  public void setXMax(double xMax) {
    this.xMax = xMax;
  }
  
  /**
   * Gets minimal output value.
   */
  public double getYMin() {
    return this.yMin;
  }
  
  /**
   * Sets minimal output value.
   */
  public void setYMin(double yMin) {
    this.yMin = yMin;
  }
  
  /**
   * Gets maximal output value.
   */
  public double getYMax() {
    return this.yMax;
  }
  
  /**
   * Sets maximal output value.
   */
  public void setYMax(double yMax) {
    this.yMax = yMax;
  }
  
  /**
   * Gets scaling type.<br>
   * true -> scaling is logarithmic instead of linear.
   */
  public boolean isLogarithmic() {
    return this.logarithmic;
  }
  
  /**
   * Sets scaling type.<br>
   * true -> scaling is logarithmic instead of linear.
   */
  public void setLogarithmic(boolean logarithmic) {
    this.logarithmic = logarithmic;
  }
  
  /**
   * Gets saturation flag.<br>
   * true -> input value out of bounds is set to minimal/maximal output value<br>
   * false -> input out of bounds is extrapolated.
   */
  public boolean isSaturated() {
    return this.saturated;
  }
  
  /** 
   * Sets saturation flag.
   */
  public void setSaturated(boolean saturated) {
    this.saturated = saturated;
  }
  
  /*
   * Computes the function value for one given value t.
   */
  public double evaluate(double x) {
    if (saturated) {      // clip value
      x = Math.max(Math.min(x, xMax), xMin);
    }
    
    double a, b, y;
    if (logarithmic) {
      a = (yMax - yMin)/Math.log(xMax/xMin);
      b = yMax - a*Math.log(xMax);
      y = a*Math.log(x) + b;
    } else {
      a = (yMax - yMin)/(xMax - xMin);
      b = yMax - a*xMax;
      y = a*x + b;
    }
    return y;
  }
  
}
