package physbeans.inout;

import javax.swing.*;
import java.awt.*;
import physbeans.math.*;
import physbeans.event.*;

/**
 * Controls for Screens consisting of a scaling panel and a display panel.<br>
 * The scaling panel allows to change the scale of a screen by sending
 * VectorEvent's with the corresponding scaling factor. Scaling can affect x or
 * y direction or both, according to the changeScale function from Screen that
 * is called. Optionally ScreenControls can provide two scale panels for
 * different scaling in x- and y-direction.<br>
 * The display panel can show x and y values that originate from a mouse click
 * to a screen and are sent via a VectorEvent.
 */
public class ScreenControls extends PhysicsPanel {

  protected boolean twoScalePanels;
  protected String scaleLabel2;
  // GUI elements
  protected NumericTextField xTextField; // display coordinates of marked point
  protected NumericTextField yTextField;
  protected JPanel positionPanel;
  protected ScalePanel scalePanel;
  protected ScalePanel scalePanel2;

  /**
   * Creates a new instance of Oscilloscope
   */
  public ScreenControls() {
    super();
    setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
    twoScalePanels = false;
    scaleLabel2 = null;
    buildGUI();
  }

  /**
   * main part of constructor as extra function to use with other screens
   *
   */
  protected void buildGUI() {
    xTextField = new NumericTextField(0.0, 5, 3, "Position:", "");
    xTextField.setEditable(false);
    yTextField = new NumericTextField(0.0, 5, 3, "", "");
    yTextField.setEditable(false);

    // one ScalePanel
    scalePanel = new ScalePanel();

    // displays for click position into positionPanel
    positionPanel = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 3));
    positionPanel.setOpaque(false);
    positionPanel.setFont(null);
    positionPanel.add(xTextField);
    positionPanel.add(yTextField);

    // controlPanel combines all controls
    add(scalePanel);
    add(Box.createHorizontalGlue());
    add(positionPanel);
  }

  /**
   * Sets label of position text fields.
   */
  public void setPositionLabel(String s) {
    xTextField.setDescription(s);
  }

  /**
   * Gets label of position text fields.
   */
  public String getPositionLabel() {
    return xTextField.getDescription();
  }

  /**
   * Sets label of scale controls.
   */
  public void setScaleLabel(String s) {
    scalePanel.setScaleLabel(s);
  }

  /**
   * Gets label of scale controls.
   */
  public String getScaleLabel() {
    return scalePanel.getScaleLabel();
  }

  /**
   * Sets initial value of the scale index.
   */
  public void setInitialScaleIndex(int newIndex) {
    scalePanel.setInitialScaleIndex(newIndex);
  }

  /**
   * Gets initial value of the scale index.
   */
  public int getInitialScaleIndex() {
    return scalePanel.getInitialScaleIndex();
  }

  /**
   * Gets last scale change. Easy access function for listeners.
   */
  public DVector getScaleChange() {
    return scalePanel.getScaleChange();
  }

  /**
   * Sets label of second scale controls.
   */
  public void setScaleLabel2(String s) {
    scaleLabel2 = s;   // store locally in case second panel is activated later
    if (scalePanel2 != null) {
      scalePanel2.setScaleLabel(s);
    }
  }

  /**
   * Gets label of second scale controls.
   */
  public String getScaleLabel2() {
    return (scalePanel2 == null) ? "" : scalePanel2.getScaleLabel();
  }

  /**
   * Sets initial value of the second scale index.
   */
  public void setInitialScaleIndex2(int newIndex) {
    if (scalePanel2 != null) {
      scalePanel2.setInitialScaleIndex(newIndex);
    }
  }

  /**
   * Gets initial value of the second scale index.
   */
  public int getInitialScaleIndex2() {
    return (scalePanel2 == null) ? -1 : scalePanel2.getInitialScaleIndex();
  }

  /**
   * Gets last scale change of second ScalePanel. Easy access function for
   * listeners.
   */
  public DVector getScaleChange2() {
    return (scalePanel2 == null) ? null : scalePanel2.getScaleChange();
  }

  /**
   * Sets text fields according to x and y coordinate of cector v.
   */
  public void setTextFields(DVector v) {
    xTextField.setValue(v.get(0));
    yTextField.setValue(v.get(1));
  }

  /**
   * Register vector listener.
   */
  public synchronized void addVectorListener(VectorListener listener) {
    scalePanel.addVectorListener(listener);
  }

  /**
   * De-register vector listener.
   */
  public synchronized void removeVectorListener(VectorListener listener) {
    scalePanel.removeVectorListener(listener);
  }

  /**
   * Register vector listener for second ScalePanel.
   */
  public synchronized void addVectorListener2(VectorListener listener) {
    if (scalePanel2 != null) {
      scalePanel2.addVectorListener(listener);
    }
  }

  /**
   * De-register vector listener for second ScalePanel.
   */
  public synchronized void removeVectorListener2(VectorListener listener) {
    if (scalePanel2 != null) {
      scalePanel2.removeVectorListener(listener);
    }
  }

  /**
   * Gets flag whether two scale panels are supplied.
   */
  public boolean isTwoScalePanels() {
    return twoScalePanels;
  }

  /**
   * Sets flag whether two scale panels are supplied.
   */
  public void setTwoScalePanels(boolean newFlag) {
    if (newFlag == twoScalePanels) {
      return;     // do nothing
    }

    if (newFlag) {
      scalePanel2 = new ScalePanel();
      add(scalePanel2, 1);
      if (scaleLabel2 != null) {
        scalePanel2.setScaleLabel(scaleLabel2);
      }
    } else {
      remove(scalePanel2);
      scalePanel2 = null;
    }
    twoScalePanels = newFlag;
  }
}
