package physbeans.inout;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;

/**
 * class that describes the transformation between screen coordinates (s,t)
 * and world coordinates (x,y)<br>
 * simple affine model:
 *    (0,0) -> (xmin, ymax)
 *    (smax, tmax) -> (xmax, ymin)
 *    (s,t) -> (x, y)<br>
 * Optionally s and t can be exchanged.
 */
public class ScreenWorldTrafo implements Serializable {

  protected Point maxScreen;  // screen size (smax, tmax)
  protected DVector minWorld;  // lower left world corner (xmin, ymin)
  protected DVector maxWorld;  // upper right world corner (xmax, ymax) 
  protected boolean constrained;   // if true, xscale = yscale
  protected boolean xyExchanged;   // if true, x is vertical, y horizontal

  // changes to screen and world sizes are rare (typically only at init)
  // transforms of coordinates are frequent
  // -> the scale factors are cached to increase transformation speed
  protected double xScale;
  protected double yScale;

  /**
   * construct default transformation
   *   screen: 100x100
   *   world: (-1,-1) .. (1,1)
   */
  public ScreenWorldTrafo() {
    maxScreen = new Point(100, 100);
    minWorld = new DVector(new double[]{-1.0, -1.0});
    maxWorld = new DVector(new double[]{1.0, 1.0});
    constrained = false;
    xyExchanged = false;

    setScales();   // compute scale factors
  }

  /**
   * construct transformation with
   *   screen: 100x100
   *   world: (xmin,ymin) .. (xmax,ymax)
   */
  public ScreenWorldTrafo(double xmin, double ymin, double xmax, double ymax) {
    maxScreen = new Point(100, 100);
    minWorld = new DVector(new double[]{xmin, ymin});
    maxWorld = new DVector(new double[]{xmax, ymax});
    constrained = false;
    xyExchanged = false;

    setScales();   // compute scale factors
  }

  /**
   * changes y coordinates so that yScale = xScale.
   */
  protected void equalizeScales() {
    if (constrained) {
      DVector factor = new DVector(1.0, xScale / yScale);
      maxWorld.arrayTimesEquals(factor);
      minWorld.arrayTimesEquals(factor);
      yScale = xScale;
    }
  }

  /**
   * Returns the ratio of x and y scaling factors.
   * This can be used to create circles or squares with non-constrained trafos:
   * Simply scale lengths in y direction with this factor.
   */
  public double getRelativeScaling() {
    return yScale / xScale;
  }

  /**
   * Returns the x scaling factor between world and pixel coordinates.
   * This can be used to create figures with dimensions defined by pixel sizes.
   */
  public double getScalingX() {
    return xScale;
  }

  /**
   * Returns the y scaling factor between world and pixel coordinates.
   * This can be used to create figures with dimensions defined by pixel sizes.
   */
  public double getScalingY() {
    return yScale;
  }

  /**
   * get maximal screen coordinates
   */
  public Point getMaxScreen() {
    return maxScreen;
  }

  /**
   * set maximal screen coordinates from given point
   */
  public void setMaxScreen(Point p) {
    // in the middle of layout the scale could be negative or 0 -> do nothing!
    if ((p.x > 0) && (p.y > 0)) {
      maxScreen.x = p.x;
      maxScreen.y = p.y;
      setScales();
    }
  }

  /**
   * set maximal screen coordinates from given dimension.<br>
   * convenience function
   */
  public void setMaxScreen(Dimension dim) {
    setMaxScreen(new Point(dim.width, dim.height));
  }

  /**
   * get maximal world coordinates
   */
  public DVector getMaxWorld() {
    return maxWorld;
  }

  /**
   * set maximal world coordinates from given vector
   */
  public void setMaxWorld(DVector v) {
    maxWorld = v.copy();
    setScales();
  }

  /**
   * get minimal world coordinates
   */
  public DVector getMinWorld() {
    return minWorld;
  }

  /**
   * set minimal world coordinates from given vector
   */
  public void setMinWorld(DVector v) {
    minWorld = v.copy();
    setScales();
  }

  /**
   * get constrained flag
   */
  public boolean isConstrained() {
    return constrained;
  }

  /**
   * set constrained flag
   */
  public void setConstrained(boolean b) {
    constrained = b;
    equalizeScales();
  }

  /**
   * Gets xyExchanged flag.
   * If true, x is vertical, y horizontal.
   */
  public boolean isXyExchanged() {
    return xyExchanged;
  }

  /**
   * Sets xyExchanged flag.
   */
  public void setXyExchanged(boolean b) {
    xyExchanged = b;
  }

  /**
   * compute scale factors for reference in transformations
   */
  protected void setScales() {
    xScale = (maxWorld.get(0) - minWorld.get(0)) / maxScreen.x;
    yScale = (maxWorld.get(1) - minWorld.get(1)) / maxScreen.y;
    equalizeScales();
  }

  /**
   * transform screen point p into world coordinates
   */
  public DVector transform(Point p) {
    double x = xScale * p.x + minWorld.get(0);
    double y = maxWorld.get(1) - yScale * p.y;

    return new DVector(new double[]{x, y});
  }

  /**
   * transform vector of screen coordinates into world coordinates
   */
  public DVector[] transform(Point[] p) {
    DVector[] v = new DVector[p.length];
    for (int i = 0; i < p.length; i++) {
      v[i] = transform(p[i]);
    }
    return v;
  }

  /**
   * transforms world vector v into screen coordinates
   */
  public Point inverseTransform(DVector v) {
    int s = (int) Math.round((v.get(0) - minWorld.get(0)) / xScale);
    int t = (int) Math.round((maxWorld.get(1) - v.get(1)) / yScale);

    return new Point(s, t);
  }

  /**
   * transforms vector of world coordinates into screen coordinates
   */
  public Point[] inverseTransform(DVector[] v) {
    Point[] p = new Point[v.length];
    for (int i = 0; i < v.length; i++) {
      p[i] = inverseTransform(v[i]);
    }
    return p;
  }

  // conveniance functions to simplify and/or accelerate
  /**
   * returns s screen coordinate for given x world coordinate
   */
  public int getS(double x) {
    return (int) Math.round((x - minWorld.get(0)) / xScale);
  }

  /**
   * returns t screen coordinate for given y world coordinate
   */
  public int getT(double y) {
    return (int) Math.round((maxWorld.get(1) - y) / yScale);
  }

  /**
   * gets maximal world x coordinate
   */
  public double getMaxX() {
    return maxWorld.get(0);
  }

  /**
   * sets maximal world x coordinate
   */
  public void setMaxX(double xMax) {
    maxWorld.set(0, xMax);
    setScales();
  }

  /**
   * gets minimal world x coordinate
   */
  public double getMinX() {
    return minWorld.get(0);
  }

  /**
   * sets minimal world x coordinate
   */
  public void setMinX(double xMin) {
    minWorld.set(0, xMin);
    setScales();
  }

  /**
   * gets maximal world y coordinate
   */
  public double getMaxY() {
    return maxWorld.get(1);
  }

  /**
   * sets maximal world y coordinate
   */
  public void setMaxY(double yMax) {
    maxWorld.set(1, yMax);
    setScales();
  }

  /**
   * gets minimal world y coordinate
   */
  public double getMinY() {
    return minWorld.get(1);
  }

  /**
   * sets minimal world y coordinate
   */
  public void setMinY(double yMin) {
    minWorld.set(1, yMin);
    setScales();
  }

  /**
   * gets maximal screen s coordinate
   */
  public int getMaxS() {
    return maxScreen.x;
  }

  /**
   * sets maximal screen s coordinate
   */
  public void setMaxS(int sMax) {
    maxScreen.x = sMax;
    setScales();
  }

  /**
   * gets maximal screen t coordinate
   */
  public int getMaxT() {
    return maxScreen.y;
  }

  /**
   * sets maximal screen t coordinate
   */
  public void setMaxT(int tMax) {
    maxScreen.y = tMax;
    setScales();
  }

  /**
   * Returns an AffineTransform trafo that corresponds to the ScreenWorldTrafo.
   * With a Graphics2D object g2 one simply sets
   *    g2.setTransform(trafo)
   * and can work with 2d drawing routines in world coordinates.
   */
  public AffineTransform getAffineTransform() {
    if (xyExchanged) {
      return new AffineTransform(0.0, -1.0 / yScale, 1.0 / xScale, 0.0,
              -minWorld.x() / xScale, maxWorld.y() / yScale);
    } else {
      return new AffineTransform(1.0 / xScale, 0.0, 0.0, -1.0 / yScale,
              -minWorld.x() / xScale, maxWorld.y() / yScale);
    }
  }

  /**
   * Returns an AffineTransform trafo that corresponds to the inverse
   * ScreenWorldTrafo.
   */
  public AffineTransform getInverseAffineTransform() {
    AffineTransform at = AffineTransform.getScaleInstance(xScale, -yScale);
    AffineTransform at1 = AffineTransform.getTranslateInstance(minWorld.x(), maxWorld.y());

    if (xyExchanged) {
      AffineTransform at2 = new AffineTransform(0.0, 1.0, 1.0, 0.0, 0.0, 0.0);
      at.concatenate(at2);
    }

    at1.concatenate(at);
    return at1;
  }
}
