package physbeans.inout;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.beans.*;
import java.io.*;
import physbeans.model.*;

/**
 * Bean for input of a numerical value via a slider or a text field
 * slider characteristic can be linear or logarithmic<br>
 * slider can send PropertyChange events either continuously (during drag) or
 *   only after drag stopped<br>
 * internal model consists of three doubles minVal <= value <= maxVal<br>
 * changes to model parameters leading to an inconsistent state are not
 *  simply rejected, but lead to a rearrangement of all three values.
 *  This should allow for simpler changes of the internal model.
 */
public class TextSlider extends PhysicsPanel implements Serializable {
  
  protected NumericTextField textField;
  protected Slider slider;
  
  /**
   * Creates a new instance of TextSlider.
   */
  public TextSlider() {
    super();
    setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
    
    textField = new NumericTextField();
    textField.setPrecision(4);   // adapted to slider precision
    textField.setColumns(5);
    textField.addPropertyChangeListener(new MyTextFieldListener());
    
    slider = new Slider();
    slider.addPropertyChangeListener(new MySliderListener());
    textField.setValueInternally(slider.getValue());
    
    // create GUI
    add(textField);
    add(slider);
    setBordered(true);
  }
  
  /**
   * Returns current value.
   */
  public double getValue() {
    return slider.getValue();
  }
  
  /**
   * Sets the current value and adjusts the JSlider accordingly.<br>
   * Fires a property change event.
   */
  public void setValue(double val) {
    setValueInternally(val);
    firePropertyChange("value", null, new Double(val));
  }
  
  /**
   * sets the current value and adjusts the JSlider accordingly
   * without firing an event.
   */
  public void setValueInternally(double val) {
    slider.setValueInternally(val);
    textField.setValueInternally(val);
  }
  
  /**
   * Listener to Slider events<br>
   *  sets text field and fires PropertyChange event
   */
  protected class MySliderListener implements PropertyChangeListener {
    public void propertyChange(PropertyChangeEvent e) {
      double val = slider.getValue();      // slider has new value
      setValueInternally(val);
      
      // inform listeners
      firePropertyChange("value", null, new Double(val));
    }
  }
  
  /**
   * Listener to NumericTextField events<br>
   *  sets value and slider and fires PropertyChange event
   */
  protected class MyTextFieldListener implements PropertyChangeListener {
    public void propertyChange(PropertyChangeEvent e) {
      // get values
      double val = textField.getValue();  // text field has new value
      
      // set (possibly corrected) new value
      double minVal = slider.getModel().getMinimum();
      double maxVal = slider.getModel().getMaximum();
      val = Math.max(Math.min(val, maxVal), minVal);
      setValueInternally(val);
      
      firePropertyChange("value", null, new Double(val));
    }
  }
  
  /**
   * Gets description string.
   */
  public String getDescription() {
    return textField.getDescription();
  }
  
  /**
   * Sets description string.
   */
  public void setDescription(String description) {
    textField.setDescription(description);
  }
  
  /**
   * Gets filename of icon used for the description field.
   */
  public String getDescriptionIcon() {
    return textField.getDescriptionIcon();
  }
  
  /**
   * Sets filename of icon used for the description field.
   */
  public void setDescriptionIcon(String file) {
    textField.setDescriptionIcon(file);
  }
  
  
  /**
   * Gets number of digits displayed in text field.
   */
  public int getPrecision() {
    return textField.getPrecision();
  }
  
  /**
   * Sets number of digits displayed in text field.
   */
  public void setPrecision(int precision) {
    textField.setPrecision(precision);
  }
  
  /**
   * Gets number of columns in text field.
   */
  public int getColumns() {
    return textField.getColumns();
  }
  
  /**
   * Sets number of columns in text field.
   */
  public void setColumns(int columns) {
    textField.setColumns(columns);
  }
  
  /**
   * Gets unit string.
   */
  public String getUnit() {
    return textField.getUnit();
  }
  
  /**
   * Sets unit string.
   */
  public void setUnit(String unit) {
    textField.setUnit(unit);
  }
  
  /**
   * Gets filename of icon used for the unit field.
   */
  public String getUnitIcon() {
    return textField.getUnitIcon();
  }
  
  /**
   * Sets filename of icon used for the unit field.
   */
  public void setUnitIcon(String file) {
    textField.setUnitIcon(file);
  }
  
  /**
   * True if events are sent continuously during slider drag.
   */
  public boolean isFireContinuously() {
    return slider.isFireContinuously();
  }
  
  /**
   * Sets fireContinuously flag.
   */
  public void setFireContinuously(boolean fireContinuously) {
    slider.setFireContinuously(fireContinuously);
  }
  
  /**
   * True if slider scale is logarithmic.
   */
  public boolean isLogarithmic() {
    return slider.isLogarithmic();
  }
  
  /**
   * Sets logarithmic flag.
   */
  public void setLogarithmic(boolean logarithmic) {
    slider.setLogarithmic(logarithmic);
  }
  
  /**
   * Returns slider model with minimum, maximum and current value.
   */
  public BoundedRealModel getModel() {
    return slider.getModel();
  }
  
  /**
   * Sets slider model with minimum, maximum and current value.
   */
  public void setModel(BoundedRealModel model) {
    slider.setModel(model);
    textField.setValueInternally(model.getValue());
  }
  
  /**
   * Sets foreground color.
   */
  public void setForeground(Color c) {
    super.setForeground(c);
    if (textField != null) {
      textField.setForeground(c);
    }
    if (slider != null) {
      slider.setForeground(c);
    }
  }
  
  /**
   * Gets value of inverted flag.<br>
   *  If true, maximal slider value is to the left.
   */
  public boolean getInverted() {
    return slider.getInverted();
  }
  
  /**
   * Sets value of inverted flag.<br>
   *  True means maximal slider value is to the left.
   */
  public void setInverted(boolean b) {
    slider.setInverted(b);
  }
}
