package physbeans.math;

import physbeans.model.*;
import java.io.*;

/**
 * Class for polynomials and their functions.<br>
 * The coefficients can be given as a vector of doubles, where the first element
 * belongs to the highest exponent.
 */
public class Polynomial implements Serializable {
  
  protected double[] a;
  protected int precision;
  
  /**
   * Creates a polynomial from given coefficients.
   */
  public Polynomial(double[] coeff) {
    a = coeff;
    precision = 4;
    normalize();
  }
  
  /**
   * Creates an empty polynomial of degree -1.
   */
  public Polynomial() {
    this(new double[0]);
  }
  
  /**
   * Creates a default polynomial of given degree, namely x^deg .
   */
  public Polynomial(int degree) {
    a = new double[degree + 1];
    a[0] = 1.0;
    precision = 4;
  }
  
  /**
   * Creates a polynomial from given coefficients.
   */
  public Polynomial(DVector coeff) {
    this(coeff.getArray());
  }
  
  /**
   * Gets the coefficient of x^i.
   */
  public double getCoeff(int index) {
    return this.a[getDegree() - index];
  }
  
  /**
   * Gets all coefficients of the polynomial.
   */
  public DVector getCoeff() {
    return new DVector(a);
  }
  
  /**
   * Sets the coefficient of x^i.
   */
  public void setCoeff(int index, double a) {
    this.a[getDegree() - index] = a;
  }
  
  /**
   * Sets all coefficients of the polynomial.
   */
  public void setCoeff(DVector v) {
    this.a = v.getArrayCopy();
  }
  
  /**
   * Gets degree of the polynomial.
   */
  public int getDegree() {
    return a.length - 1;
  }
  
  /**
   * Creates standard polynomial of the given degree.
   */
  public void setDegree(int degree) {
    a = new double[degree + 1];
    a[0] = 1.0;
  }
  
  /**
   * Returns a String representation of the polynomial.
   */
  public String toString() {
    int degree = getDegree();
    
    if (degree < 0) {
      return "invalid!";
    }
    
    // highest coeff without explicit sign
    String s = "" + Maths.toString(a[0], precision);
    if (degree > 0) {
      s += "*x";
    }
    if (degree > 1) {
      s += "^" + degree;
    }
    
    // next all other coeff's with explicit exponents
    for (int i = degree - 1; i > 1; i--) {
      if (a[degree - i] != 0.0) {
        s += " " + toSignedString(a[degree - i]) + "*x^" + i;
      }
    }
    
    // now linear coeff
    // degree = 1 -> already done above !
    if (degree > 1 && a[degree - 1] != 0.0) {
      s += " " + toSignedString(a[degree - 1]) + "*x";
    }
    
    // and finally constant coeff
    // degree = 0 -> already done above !
    if (degree > 0 && a[degree] != 0.0) {
      s += " " + toSignedString(a[degree]);
    }
    
    return s;
  }
  
  /**
   * Converts a double to a string with an explicit sign.
   */
  protected  String toSignedString(double d) {
    // erzeugt String mit explizitem Vorzeichen
    if (d >= 0) {
      return("+ " + Maths.toString(d, precision));
    } else {
      return("- " + Maths.toString(-d, precision));
    }
  }
  
  /**
   * Returns polynomial of given degree that fits the points in a least
   * square sense.
   */
  public static Polynomial computeFitPolynomial(DVector x, DVector y, int degree) {
    // reduce degree, if number of points is too low
    int n = x.getDimension();
    degree = Math.min(degree, n - 1);
    
    // setup coefficient matrix of normal equation
    double[][] a = new double[degree + 1][degree + 1];
    for (int i = 0; i <= degree; i++) {
      for (int j = 0; j <= degree; j++) {
        for (int l = 0; l < n; l++) {
          a[i][j] += Math.pow(x.get(l), i+j);
        }
      }
    }
    Matrix matA = new Matrix(a);
    
    // setup rhs vector of normal equation
    double[] r = new double[degree + 1];
    for (int i = 0; i <= degree; i++) {
      for (int l = 0; l < n; l++) {
        r[i] += y.get(l)*Math.pow(x.get(l), i);
      }
    }
    Matrix matR = new Matrix(r, degree + 1);
    
    // solve normal equation
    Matrix matX = matA.solve(matR);
    
    // reorder vector;
    DVector coeff = new DVector(matX.getColumnPackedCopy());
    coeff.reverse();
    return new Polynomial(coeff);
  }
  
  /**
   * Returns a vector that results of the application of the polynomial to
   * all input values.
   */
  public DVector apply(DVector inVal) {
    int size = inVal.getDimension();
    DVector outVal = new DVector(size);
    
    // loop over all vector elements
    for (int i = 0; i < size; i++) {
      double x = inVal.get(i);
      outVal.set(i, apply(x));
    }
    return outVal;
  }
  
  /**
   * Returns Polynomial(x).
   */
  public double apply(double x) {
    int degree = getDegree();
    double y;
    if (degree >= 0) {
      y = a[0];
      for (int k = 1; k <= degree; k++) {
        y = y*x + a[k];
      }
    } else {
      y = Double.NaN;
    }
    return y;
  }
  
  /**
   * Gets precision of doubles in toString.
   */
  public int getPrecision() {
    return this.precision;
  }
  
  /**
   * Sets precision of doubles in toString.
   */
  public void setPrecision(int precision) {
    this.precision = precision;
  }
  
  /**
   * Computes the (complex) roots of the polynomial and returns them
   * as a CVector.
   */
  public CVector getRoots() {
    // check polynomial
    if (a.length == 0) {
      throw new ArithmeticException("Polynomial = 0!");
    }
    if (a.length == 1) {
      throw new ArithmeticException("Constant Polynomial!");
    }
    if (a[a.length-1] == 0.0) {
      // shouldn't happen
      normalize();
    }
    
    // construct the companion matrix
    int dim = a.length - 1;
    Matrix m = new Matrix(dim, dim);
    for (int i = 1; i <dim; i++) {
      // set lower subdiagonal to 1
      m.set(i, i-1, 1.0);
    }
    for (int i = 0; i < dim; i++) {
      // set first line to normalized coefficients
      m.set(0, i, -a[i+1]/a[0]);
    }
    
    // get eigenvalues of m
    EigenvalueDecomposition ev = new EigenvalueDecomposition(m);
    
    // build result array
    CVector result = new CVector(ev.getRealEigenvalues(),
        ev.getImagEigenvalues());
    return result;
  }
  
  /**
   * gets rid of leading zero coefficients
   */
  protected void normalize() {
    /*
    int i;
    for (i = a.length-1; i >=0; i--) {
      if (a[i] != 0.0) break;
      if (i == 0) {  // zero polynomial
        a = new double[0];
        return;
      }
    }
     
    if (i < a.length) {
      // leading zeroes found -> copy into smaller array
      double[] b = new double[i+1];
      for (int k = 0; k < i+1; k++) {
        b[k] = a[k];
        a = b;
      }
    }   */
  }
  
}
