package physbeans.phys;

/**
 * A vertical wall with a number of equally sized and equally spaced slits,
 * i.e. a grating.
 */
public class GratingWall extends WaveWall {

  protected int n;
  protected double d, g;
  protected double pos;
  protected int thickness;

  /**
   * Creates a new instance of GratingWall
   */
  public GratingWall() {
    super();
    n = 2;
    d = 0.4;
    g = 0.8;
    pos = -0.9;
    thickness = 1;
  }

  protected void setGrid() {
    if ((field == null) || (grid == null)) {
      return;
    }

    initGrid();
    int nx = field.getGridSizeX();
    int ny = field.getGridSizeY();

    // create wall of given thickness
    int px1 = field.getIndexForX(pos);     // left end of grating
    if (px1 == -1) {     // grating lies outside field grid
      return;
    }
    int px2 = Math.min(px1 + thickness, nx);

    for (int i = 0; i < ny; i++) {
      for (int px = px1; px < px2; px++) {
        grid[i][px] = 1.0;
      }
    }

    // create slits
    for (int i = 0; i < n; i++) {
      double y = g * (i - (n - 1) / 2.0);    // center of slit i
      int py1 = field.getIndexForY(y - d / 2);
      int py2 = field.getIndexForY(y + d / 2);
      if ((py1 == -1) && (py2 == -1)) {
        // slit is outside or spreads the whole wall, which is useless
        continue;
      } else {        // care for possible partial slits
        py2 = (py2 == -1) ? 0 : py2;
        py1 = (py1 == -1) ? ny - 1 : py1;
      }

      for (int j = py2; j < py1; j++) {
        for (int px = px1; px < px2; px++) {
          grid[j][px] = 0.0;
        }
      }
    }

    trigger();
  }

  /**
   * Gets the number of slits.
   */
  public int getNumberOfSlits() {
    return n;
  }

  /**
   * Sets the number of slits.
   */
  public void setNumberOfSlits(int n1) {
    if (n1 >= 0) {
      n = n1;
      setGrid();
    }
  }

  /**
   * Gets the width of the slits.
   */
  public double getSlitWidth() {
    return d;
  }

  /**
   * Sets the width of the slits.
   */
  public void setSlitWidth(double d1) {
    if (d1 > 0) {
      d = d1;
      setGrid();
    }
  }

  /**
   * Gets the thickness of the grating in pixels.
   */
  public int getThickness() {
    return thickness;
  }

  /**
   * Sets the thickness of the grating in pixels.
   */
  public void setThickness(int th) {
    if (th > 0) {
      thickness = th;
      setGrid();
    }
  }

  /**
   * Gets the spacing of the slits.
   */
  public double getSpacing() {
    return g;
  }

  /**
   * Sets the spacing of the slits.
   */
  public void setSpacing(double g1) {
    if (g1 > 0) {
      g = g1;
      setGrid();
    }
  }

  /**
   * Gets the position (i.e. the x coordinate) of the grating.
   */
  public double getPosition() {
    return pos;
  }

  /**
   * Sets the position (i.e. the x coordinate) of the grating.
   */
  public void setPosition(double pos1) {
    pos = pos1;
    setGrid();
  }
}
