package physbeans.phys;

import physbeans.math.*;

/**
 * Model of an ideal gas with constant heat capacity. <br>
 * Describes material properties (M, kappa) and current state of the gas
 *  (m, p, V, T).
 * To define the caloric state variables (U, H, S) the values U0 and S0 at the
 * standard conditions p0, T0 can be set arbitrarily.<br>
 * The gas reacts according to state changes and has bookkeeping variables
 * for the amount of transferred heat and work.
 */
public class PerfectGas extends ThermodynamicSystem {
  
  protected double p;
  protected double V;
  
  protected double m;
  protected double M;
  protected double kappa;
  
  protected double Ri;  // ok, it's redundant. But: It's much clearer!
  protected double p0;  // standard conditions
  protected double T0;
  protected double V0;
  
  protected double U0;
  protected double S0;
  
  protected boolean autoReset;
  protected double WV;
  protected double Wt;
  protected double Q;
  
  /**
   * Creates a default instance of PerfectGas.
   * Material constants are those of air at standard conditions.
   */
  public PerfectGas() {
    super();
    
    m = 1.0;
    M = 28.953e-3;
    Ri = Constants.R/M;
    kappa = 1.40;
    
    p0 = 1.01325e5;
    T0 = Constants.T0K;
    V0 = m*Ri*T0/p0;
    
    U0 = 0.0;
    S0 = 0.0;
    autoReset = true;
    
    p = p0;
    V = V0;
    
    restart();
  }
  
  /**
   * Gets current pressure p.
   * @return Value of pressure p in Pa.
   */
  public double getPressure() {
    return this.p;
  }
  
  /**
   * Sets current pressure p.
   * @param p New value of pressure p in Pa.
   */
  public void setPressure(double p) {
    setPressureInternally(p);
  }
  
  /**
   * Sets current pressure p without triggering.
   * @param p New value of pressure p in Pa.
   */
  public void setPressureInternally(double p) {
    if (p <= 0) {
      throw new IllegalArgumentException("non-positive pressure");
    }
    this.p = p;
  }
  
  /**
   * Gets current volume V.
   * @return Value of volume V in m^3.
   */
  public double getVolume() {
    return this.V;
  }
  
  /**
   * Sets current volume V.
   * @param V New value of volume V.
   */
  public void setVolume(double V) {
    setVolumeInternally(V);
  }
  
  /**
   * Sets current volume V without triggering.
   * @param V New value of volume V.
   */
  protected void setVolumeInternally(double V) {
    if (V <= 0) {
      throw new IllegalArgumentException("non-positive volume");
    }
    this.V = V;
  }
  
  /**
   * Gets gas mass m.
   * @return Value of mass m in kg.
   */
  public double getMass() {
    return this.m;
  }
  
  /**
   * Sets gas mass m.
   * @param m New value of mass m in kg.
   */
  public void setMass(double m) {
    this.m = m;
    V0 = m*Ri*T0/p0;   // change volume at norm conditions
  }
  
  /**
   * Gets molar mass M.
   * @return Value of molar Mass M in kg/mol.
   */
  public double getMolarMass() {
    return this.M;
  }
  
  /**
   * Sets molar mass M.
   * @param M New value of molar Mass M in kg/mol.
   */
  public void setMolarMass(double M) {
    this.M = M;
    Ri = Constants.R/M;
  }
  
  /**
   * Gets current temperature T in K.
   */
  public double getTemperature() {
    return p*V/(m*Ri);
  }
  
  /**
   * Gets inner energy U0 in J at standard conditions.
   * @return Value of U0 in J.
   */
  public double getU0() {
    return this.U0;
  }
  
  /**
   * Sets inner energy U0 in J at standard conditions.
   * This fixes the value of the inner energy for all conditions and can be set
   * arbitrarily.
   * @param U0 New value of U0 in J.
   */
  public void setU0(double U0) {
    this.U0 = U0;
  }
  
  /**
   * Gets entropy S0 in J/K at standard conditions.
   * @return Value of S0 in J/K.
   */
  public double getS0() {
    return this.S0;
  }
  
  /**
   * Sets entropy S0 in J/K at standard conditions.
   * This fixes the value of the entropy for all conditions and can be set
   * arbitrarily.
   * @param S0 New value of S0 in J/K.
   */
  public void setS0(double S0) {
    this.S0 = S0;
  }
  
  /**
   * Gets adiabatic exponent kappa.
   * @return Value of kappa.
   */
  public double getKappa() {
    return this.kappa;
  }
  
  /**
   * Sets adiabatic exponent kappa.
   * @param kappa New value of kappa.
   */
  public void setKappa(double kappa) {
    this.kappa = kappa;
  }
  
  /**
   * Gets value of the heat capacity at constant pressure c_p in J/(kg K).
   */
  public double getCP() {
    return Ri*kappa/(kappa - 1.0);
  }
  
  /**
   * Gets value of the heat capacity at constant volume c_V in J/(kg K).
   */
  public double getCV() {
    return Ri/(kappa - 1.0);
  }
  
  /**
   * Gets current value of the inner energy U in J.
   * Reference point is U0 at T0.
   */
  public double getInnerEnergy() {
    double T = getTemperature();
    double U = m*getCV()*(T - T0) + U0;
    return U;
  }
  
  /**
   * Gets current value of the enthalpy H in J.
   * Reference point is U0 at T0.
   */
  public double getEnthalpy() {
    double H = getInnerEnergy() + p*V;
    return H;
  }
  
  /**
   * Gets current value of the entropy S in J/K.
   * Reference point is S0 at norm conditions.
   */
  public double getEntropy() {
    double S = m*getCV()*(Math.log(p/p0) + kappa*Math.log(V/V0)) + S0;
    return S;
  }
  
  /**
   * Gets current accumulated value of the work WV in J done by volume changes.
   * Value is a sum over all state changes since last restart.
   */
  public double getWork() {
    return WV;
  }
  
  /**
   * Gets current accumulated value of the technical work Wt in J.
   * Value is a sum over all state changes since last restart.
   */
  public double getTechnicalWork() {
    return Wt;
  }
  
  /**
   * Gets current accumulated value of the heat in J.
   * Value is a sum over all state changes since last restart.
   */
  public double getHeat() {
    return Q;
  }
  
  /**
   * Sets value of the accumulated work, technical work and heat to 0.
   */
  public void restart() {
    WV = 0.0;
    Wt = 0.0;
    Q = 0.0;
    
    tryTrigger();
  }
  
  /**
   * Changes the state of the gas according to a polytrope of exponent n.
   * The endpoint is defined by the type of variable given (P, V or T) and
   * its final value.<br>
   * Adds the amount of work and heat that are transferred during this state change
   * to the internal accumulation variables.<br>
   * Some typical values for the exponent n:
   * <ul>
   * <li>n = 0: isobaric</li>
   * <li>n = 1: isothermal</li>
   * <li>n = kappa: isentropic</li>
   * <li>n = infinity: isochoric</li>
   * </ul>
   */
  public void changeState(double n, ThermodynamicStateVariable tsv, double val) {
    double pNew, VNew;
    
    if (n == Double.POSITIVE_INFINITY) {
      isochoricStateChange(tsv, val);
      return;
    }
    
    // compute pNew, VNew
    if (tsv == ThermodynamicStateVariable.V) {
      VNew = val;
      pNew = p * Math.pow(V/VNew, n);
    } else if (tsv == ThermodynamicStateVariable.P) {
      if (Maths.isEqual(n, 0.0)) {
        throw new IllegalArgumentException("trying to define end pressure for an isobare");
      }
      pNew = val;
      VNew = V * Math.pow(p/pNew, 1.0/n);
    } else if (tsv == ThermodynamicStateVariable.T) {
      if (Maths.isEqual(n, 1.0)) {
        throw new IllegalArgumentException("trying to define end temperature for an isothermal");
      }
      double T = getTemperature();
      pNew = p * Math.pow(val/T, n/(n - 1.0));
      VNew = V * Math.pow(T/val, 1.0/(n - 1.0));
    } else {
      throw new IllegalArgumentException("wrong value of a ThermodynamicStateVariable");
    }
    
    // compute works and heat
    double dWV;
    if (Maths.isEqual(n, 1.0)) {
      dWV = p*V*Math.log(V/VNew);
    } else {
      dWV = p*V/(n - 1.0) * (Math.pow(V/VNew, n - 1.0) - 1.0);
    }
    WV += dWV;
    Wt += n*dWV;
    Q  += (n - kappa)/(kappa - 1.0) * dWV;
    
    // set state variables
    p = pNew;
    V = VNew;
    
    tryTrigger();
  }
  
  /**
   * Changes state according to an isochore.
   * Special case of the routine changeState.
   */
  protected void isochoricStateChange(ThermodynamicStateVariable tsv, double val) {
    double pNew;
    if (tsv == ThermodynamicStateVariable.V) {
      throw new IllegalArgumentException("trying to define end volume for an isochore");
    } else if (tsv == ThermodynamicStateVariable.P) {
      pNew = val;
    } else if (tsv == ThermodynamicStateVariable.T) {
      pNew = p * val/getTemperature();
    } else {
      throw new IllegalArgumentException("wrong value of a ThermodynamicStateVariable");
    }
    
    
    // WV is unchanged
    double dWt = V*(pNew - p);
    Wt += dWt;
    Q += dWt/(kappa - 1);
    
    p = pNew;
    
    tryTrigger();
  }
  
  /**
   * Gets the current value of a thermodynamic state variable of type tsv.
   */
  public double getStateValue(ThermodynamicStateVariable tsv) {
    if (tsv == ThermodynamicStateVariable.V) {
      return V;
    } else if (tsv == ThermodynamicStateVariable.P) {
      return p;
    } else if (tsv == ThermodynamicStateVariable.T) {
      return getTemperature();
    } else {
      throw new IllegalArgumentException("wrong value of a ThermodynamicStateVariable");
    }
  }
  
  /**
   * Sets an arbitrary (initial) state given by (p, V).
   * Initialises the internal energy accumulators.
   */
  public void setState(double pi, double Vi) {
    if (pi > 0.0 && Vi > 0.0) {
      setVolumeInternally(Vi);
      setPressureInternally(pi);
      if (autoReset) {
        restart();    // includes trigger()!
      } else {
        tryTrigger();
      }
    }
  }
  
  /**
   * Gets value of autoReset flag.
   * If true (its default) the setState method resets the energy accumulators.
   */
  public boolean isAutoReset() {
    
    return this.autoReset;
  }
  
  /**
   * Sets value of autoReset flag.
   * If true (its default) the setState method resets the energy accumulators.
   */
  public void setAutoReset(boolean autoReset) {
    
    this.autoReset = autoReset;
  }
}
