package physbeans.phys;

import static java.lang.Math.*;
import physbeans.math.*;
import physbeans.model.*;
import static physbeans.math.Maths.*;
import static physbeans.math.Complex.*;

/**
 *
 * Schroedinger2DPML simulates the two-dimensional Schrödinger equation on a
 * rectangular domain surrounded by a non-reflecting boundary.
 * The PML rises quadratically with distance from the interior according to
 *     d(x) = 1/(1 + i sigma0 x^2)
 */
public class Schroedinger2DPML extends Schroedinger2D {

  protected CMatrix psi;
  protected CMatrix damp;
  protected double sigma0;
  // additional cached values
  protected double al[];
  protected CMatrix caCache[][], saCache[][], a1Cache[], b1Cache[], VexpCache[];
  protected FastCMatrix chi1, chi2, phi, ca1, sa1, a11, b11, psi1, psi2;

  /** Creates a new instance of Schroedinger2D */
  public Schroedinger2DPML() {
    super();
    sigma0 = 1.2;
  }

  /**
   * Computes all auxiliary values for a change of the grid
   * structure.
   */
  protected void setupGrid() {
    super.setupGrid();
  }

  /**
   * Gets sigma0 parameter of the quadratic PML layer.
   */
  public double getSigma0() {
    return sigma0;
  }

  /**
   * Sets sigma0 parameter of the quadratic PML layer.
   */
  public void setSigma0(double s0) {
    sigma0 = s0;
    setupGrid();
  }

  /**
   * Computes  psi = U1(dt) psi
   * for the first-order approximant U1 of exp(-i*H*dt/hbar).
   */
  protected void applyU1(int dtIndex) {
    IndexRange idx, idy;
    int sx, sy;

    // exp(-i*H1*dt/hbar)
    idx = new IndexRange(2, 4, nx - 5);
    idy = new IndexRange(2, ny - 3);
    sx = 2;
    sy = 0;
    applyUalpha(0, dtIndex, idx, idy, sx, sy, false);
    idx = new IndexRange(3, 4, nx - 5);
    applyUalpha(0, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H2*dt/hbar)
    idx = new IndexRange(4, 4, nx - 5);
    idy = new IndexRange(2, ny - 3);
    sx = 2;
    sy = 0;
    applyUalpha(0, dtIndex, idx, idy, sx, sy, false);
    idx = new IndexRange(5, 4, nx - 5);
    applyUalpha(0, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H3*dt/hbar)
    idx = new IndexRange(2, 2, nx - 4);
    idy = new IndexRange(2, ny - 3);
    sx = 1;
    sy = 0;
    applyUalpha(1, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H4*dt/hbar)
    idx = new IndexRange(3, 2, nx - 4);
    idy = new IndexRange(2, ny - 3);
    sx = 1;
    sy = 0;
    applyUalpha(1, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H5*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(2, 4, ny - 5);
    sx = 0;
    sy = 2;
    applyUalpha(2, dtIndex, idx, idy, sx, sy, false);
    idy = new IndexRange(3, 4, ny - 5);
    applyUalpha(2, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H6*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(4, 4, ny - 5);
    sx = 0;
    sy = 2;
    applyUalpha(2, dtIndex, idx, idy, sx, sy, false);
    idy = new IndexRange(5, 4, ny - 5);
    applyUalpha(2, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H7*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(2, 2, ny - 4);
    sx = 0;
    sy = 1;
    applyUalpha(3, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H8*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(3, 2, ny - 4);
    sx = 0;
    sy = 1;
    applyUalpha(3, dtIndex, idx, idy, sx, sy, false);

    // exp(-i*H9*dt/hbar)
    CMatrix Vexp = VexpCache[dtIndex];
    psi.arrayTimesEquals(Vexp);
  }

  /**
   * Computes  psi = U1^T(dt) psi
   * for the transpose of the first-order approximant U1 of exp(-i*H*dt/hbar).
   */
  protected void applyU1T(int dtIndex) {
    IndexRange idx, idy;
    int sx, sy;

    // exp(-i*H9*dt/hbar)
    CMatrix Vexp = VexpCache[dtIndex];
    psi.arrayTimesEquals(Vexp);

    // exp(-i*H8*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(3, 2, ny - 4);
    sx = 0;
    sy = 1;
    applyUalpha(3, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H7*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(2, 2, ny - 4);
    sx = 0;
    sy = 1;
    applyUalpha(3, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H6*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(4, 4, ny - 5);
    sx = 0;
    sy = 2;
    applyUalpha(2, dtIndex, idx, idy, sx, sy, true);
    idy = new IndexRange(5, 4, ny - 5);
    applyUalpha(2, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H5*dt/hbar)
    idx = new IndexRange(2, nx - 3);
    idy = new IndexRange(2, 4, ny - 5);
    sx = 0;
    sy = 2;
    applyUalpha(2, dtIndex, idx, idy, sx, sy, true);
    idy = new IndexRange(3, 4, ny - 5);
    applyUalpha(2, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H4*dt/hbar)
    idx = new IndexRange(3, 2, nx - 4);
    idy = new IndexRange(2, ny - 3);
    sx = 1;
    sy = 0;
    applyUalpha(1, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H3*dt/hbar)
    idx = new IndexRange(2, 2, nx - 4);
    idy = new IndexRange(2, ny - 3);
    sx = 1;
    sy = 0;
    applyUalpha(1, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H2*dt/hbar)
    idx = new IndexRange(4, 4, nx - 5);
    idy = new IndexRange(2, ny - 3);
    sx = 2;
    sy = 0;
    applyUalpha(0, dtIndex, idx, idy, sx, sy, true);
    idx = new IndexRange(5, 4, nx - 5);
    applyUalpha(0, dtIndex, idx, idy, sx, sy, true);

    // exp(-i*H1*dt/hbar)
    idx = new IndexRange(2, 4, nx - 5);
    idy = new IndexRange(2, ny - 3);
    sx = 2;
    sy = 0;
    applyUalpha(0, dtIndex, idx, idy, sx, sy, true);
    idx = new IndexRange(3, 4, nx - 5);
    applyUalpha(0, dtIndex, idx, idy, sx, sy, true);
  }

  /**
   * Computes psi = Ualpha(dt) psi
   * where Ualpha is one of the (many) 2x2 matrices which make up U1.
   *
   * @param alIndex     index of value alpha
   * @param dtIndex     index of value dt
   * @param idx,idy     basic index values U works on
   * @param sx,sy       distance to second indices for U
   * @param trans       if true, compute transpose matrix
   */
  protected void applyUalpha(int alIndex, int dtIndex,
          IndexRange idx, IndexRange idy, int sx, int sy, boolean trans) {

    CMatrix a1, b1;
    if (trans) {
      b1 = a1Cache[alIndex];
      a1 = b1Cache[alIndex];
    } else {
      a1 = a1Cache[alIndex];
      b1 = b1Cache[alIndex];

    }
    CMatrix ca = caCache[alIndex][dtIndex];
    CMatrix sa = saCache[alIndex][dtIndex];

    int subx = idx.getSize();
    int suby = idy.getSize();
    ca1.resize(suby, subx);
    sa1.resize(suby, subx);
    a11.resize(suby, subx);
    b11.resize(suby, subx);
    psi1.resize(suby, subx);
    psi2.resize(suby, subx);
    chi1.resize(suby, subx);
    chi2.resize(suby, subx);
    phi.resize(suby, subx);

    ca1.setMatrix(idy, idx, ca);
    sa1.setMatrix(idy, idx, sa);
    a11.setMatrix(idy, idx, a1);
    b11.setMatrix(idy, idx, b1);
    psi1.setMatrix(idy, idx, psi);
    psi2.setMatrix(idy.add(sy), idx.add(sx), psi);

    chi1.setArrayTimes(ca1, psi1);
    phi.setArrayTimes(a11, sa1);
    phi.arrayTimesEquals(psi2).timesEquals(Complex.I);
    chi1.plusEquals(phi);

    chi2.setArrayTimes(ca1, psi2);
    phi.setArrayTimes(b11, sa1);
    phi.arrayTimesEquals(psi1).timesEquals(Complex.I);
    chi2.plusEquals(phi);

    psi.setSubMatrix(idy, idx, chi1);
    psi.setSubMatrix(idy.add(sy), idx.add(sx), chi2);
  }

  /**
   * Creates the damping matrix implementing the PML.
   */
  protected void createAbsorbingBoundary() {
    // create arrays if necessary
    if ((damp == null) || (damp.getRowDimension() != ny) || (damp.getColumnDimension()
            != nx)) {
      damp = new CMatrix(ny, nx);
    }

// interior points are 1
    for (int i = bwX; i < nx - bwX; i++) {
      for (int j = bwY; j < ny - bwY; j++) {
        damp.set(j, i, new Complex(1.0));
      }

    }

    // careful for bwX != bwY!
    // sigma0 is the maximal value at the larger boundary
    int bw = min(bwX, bwY);

    // fill boundary rectangles starting from the inside
    for (int i = 0; i < bw; i++) {
      double sigma = sigma0 / (sqr(max(bwX, bwY))) * sqr(i + 1);
      Complex d = over(1.0, plus(1.0, new Complex(0, sigma))); // = 1/(1+i sigma)

      for (int j = bwX - 1 - i; j < nx - bwX + i + 1; j++) {
        damp.set(bwY - 1 - i, j, d);
        damp.set(ny - bwY + i, j, d);
      }

      for (int j = bwY - i; j < ny - bwY + i; j++) {
        damp.set(j, bwX - 1 - i, d);
        damp.set(j, nx - bwX + i, d);
      }

    }

    // now care for extra boundary in one dimension
    if (bwX > bw) {
      for (int i = bw; i < bwX; i++) {
        double sigma = sigma0 / (sqr(max(bwX, bwY))) * sqr(i + 1);
        Complex d = over(1.0, plus(1.0, new Complex(0, sigma)));
        for (int j = 0; j < ny; j++) {
          damp.set(j, bwX - 1 - i, d);
          damp.set(j, nx - bwX + i, d);
        }

      }
    } else if (bwY > bw) {
      for (int i = bw; i < bwY; i++) {
        double sigma = sigma0 / (sqr(max(bwX, bwY))) * sqr(i + 1);
        Complex d = over(1.0, plus(1.0, new Complex(0, sigma)));
        for (int j = 0; j < nx; j++) {
          damp.set(bwY - 1 - i, j, d);
          damp.set(ny - bwY + i, j, d);
        }

      }
    }
  }

  /**
   * Pre-computes matrices to speed up the time step.
   */
  protected void computeCachedMatrices() {
    createAbsorbingBoundary();

    p = 1.0 / (4.0 - pow(4.0, 1.0 / 3.0));  // time step parameter in U2

    double dx = getDX();
    double dy = getDY();
    double al0 = 2 * hbar / (me * (dx * dx + dy * dy));
    double dt = getTimeIncrement();

    tStep = new double[]{0.5 * p * dt, 0.5 * (1 - 4 * p) * dt};
    al = new double[]{-al0 / 96, al0 / 6, -al0 / 96, al0 / 6, -al0 / 24};

    // create arrays if necessary
    if (caCache == null) {
      caCache = new CMatrix[4][2];
      saCache = new CMatrix[4][2];
      a1Cache = new CMatrix[4];
      b1Cache = new CMatrix[4];
      VexpCache = new CMatrix[2];
    }

    int suby = ny / 2;        // reasonable start values
    int subx = nx / 2;
    chi1 = new FastCMatrix(suby, subx);
    chi2 = new FastCMatrix(suby, subx);
    phi = new FastCMatrix(suby, subx);
    ca1 = new FastCMatrix(suby, subx);
    sa1 = new FastCMatrix(suby, subx);
    a11 = new FastCMatrix(suby, subx);
    b11 = new FastCMatrix(suby, subx);
    psi1 = new FastCMatrix(suby, subx);
    psi2 = new FastCMatrix(suby, subx);

    // matrices for H1/H2
    int i = 0;
    CMatrix a = damp.arrayTimes(damp.shift(0, -2).times(3.0).plus(damp.shift(0, 2)));
    CMatrix b = damp.shift(0, -2).arrayTimes(damp.shift(0, -4).plus(damp.times(3.0)));
    CMatrix sqrtab = a.arrayTimes(b).sqrt();

    a1Cache[i] = a.arrayRightDivide(sqrtab);
    b1Cache[i] = b.arrayRightDivide(sqrtab);
    for (int j = 0; j < 2; j++) {
      caCache[i][j] = sqrtab.times(al[i] * tStep[j]).cos();
      saCache[i][j] = sqrtab.times(al[i] * tStep[j]).sin();
    }

    // matrices for H3/H4
    i = 1;
    a = damp.arrayTimes(damp.shift(0, -1).times(3.0).plus(damp.shift(0, 1)));
    b = damp.shift(0, -1).arrayTimes(damp.shift(0, -2).plus(damp.times(3.0)));
    sqrtab = a.arrayTimes(b).sqrt();

    a1Cache[i] = a.arrayRightDivide(sqrtab);
    b1Cache[i] = b.arrayRightDivide(sqrtab);
    for (int j = 0; j < 2; j++) {
      caCache[i][j] = sqrtab.times(al[i] * tStep[j]).cos();
      saCache[i][j] = sqrtab.times(al[i] * tStep[j]).sin();
    }

    // matrices for H5/H6
    i = 2;
    a = damp.arrayTimes(damp.shift(-2, 0).times(3.0).plus(damp.shift(2, 0)));
    b = damp.shift(-2, 0).arrayTimes(damp.shift(-4, 0).plus(damp.times(3.0)));
    sqrtab = a.arrayTimes(b).sqrt();

    a1Cache[i] = a.arrayRightDivide(sqrtab);
    b1Cache[i] = b.arrayRightDivide(sqrtab);
    for (int j = 0; j < 2; j++) {
      caCache[i][j] = sqrtab.times(al[i] * tStep[j]).cos();
      saCache[i][j] = sqrtab.times(al[i] * tStep[j]).sin();
    }

    // matrices for H7/H8
    i = 3;
    a = damp.arrayTimes(damp.shift(-1, 0).times(3.0).plus(damp.shift(1, 0)));
    b = damp.shift(-1, 0).arrayTimes(damp.shift(-2, 0).plus(damp.times(3.0)));
    sqrtab = a.arrayTimes(b).sqrt();

    a1Cache[i] = a.arrayRightDivide(sqrtab);
    b1Cache[i] = b.arrayRightDivide(sqrtab);
    for (int j = 0; j < 2; j++) {
      caCache[i][j] = sqrtab.times(al[i] * tStep[j]).cos();
      saCache[i][j] = sqrtab.times(al[i] * tStep[j]).sin();
    }

    // matrices for H9
    i = 4;
    a = damp.times(-1.0).arrayTimes(
            damp.shift(0, -2).plus(damp.shift(0, -1).times(-16.0)).plus(
            damp.shift(0, 1).times(-16.0)).plus(damp.shift(0, 2)).plus(
            damp.shift(-2, 0)).plus(damp.shift(-1, 0).times(-16.0)).plus(
            damp.shift(1, 0).times(-16.0)).plus(damp.shift(2, 0)));

    // potential function
    for (int j = 0; j < 2; j++) {
      VexpCache[j] = new CMatrix(ny, nx, new Complex(1.0, 0.0));
    }

    if (V != null) {
      gridV = V.getGrid();
      if (gridV == null) {
        return;
      }
      CMatrix Vxy = new CMatrix(gridV);

      for (int j = 0; j < 2; j++) {
        Complex fac = Complex.I.times(-tStep[j] / hbar);
        VexpCache[j] = Vxy.times(fac).exp();
      }
    }

    for (int j = 0; j < 2; j++) {
      CMatrix ca1 = a.times(al[i] * tStep[j]).cos();
      CMatrix sa1 = a.times(al[i] * tStep[j]).sin().timesEquals(Complex.I);
      ca1.plusEquals(sa1);
      VexpCache[j].arrayTimesEquals(ca1);
    }

    psi = new CMatrix(val, valI);    // ref's, no copies!

  }
}
