package physbeans.phys;

import physbeans.model.*;
import physbeans.math.*;

/**
 * Model of a spherical mirror using the correct reflection law or the paraxial
 * approximation. In the latter case the actual reflexion is computed at the
 * tangential plane perpendicular to the optical axis ("principal plane"). 
 * <P>The mirror reflects at both sides and can be made convex or concave to the 
 * left (or right) by choosing the sign of the curvature radius.
 */
public class SphericalMirror extends ImagingElement {
  
  protected double r;
  protected double h;
  protected boolean paraxialApproximation;
  
  /** Creates a new instance of SimpleSphericalMirror */
  public SphericalMirror() {
    super();
    r = -5.0;
    h = 2.0;
    paraxialApproximation = true;
  }
  
  protected void extend(LightRay ray) {
    DVector pLast = ray.getLastPoint();
    double al = ray.getAngle();
    
    if (paraxialApproximation) {
      if (rayIsDeparting(ray)) {
        return;
      }
      
      double y1 = getHittingHeight(pLast, al);
      if (Math.abs(y1) > h) {
        return;         // miss
      }
      ray.addPoint(pos, y1);
      ray.setAngle(getApproximateReflectionAngle(pLast, y1));
    } else {
      Circle2D circle = new Circle2D(new DVector(pos + r, 0.0), r, 
          getLeftEdge(), getRightEdge());
      Ray2D endRay = new Ray2D(pLast, al);
      DVector pHit = circle.getSegmentRayHitPoint(endRay);
      
      if (Double.isNaN(pHit.x())) {
        return;    // miss
      }
      ray.addPoint(pHit);
      
      DVector in = new DVector(Math.cos(al), Math.sin(al));
      DVector n = circle.getNormal(pHit);
      
      ray.setAngle(OpticalSurface.getReflectionAngle(in, n));
      propagate(ray);
    }
  }
  
  /**
   * Gets the radius of the curvature.
   * A positive value means the center of the sphere lies to the right of
   * the mirror. This is different from the usual mirror conventions, but in
   * accordance with standard lens conventions.
   */
  public double getRadius() {
    return this.r;
  }
  
  /**
   * Sets the radius of the curvature.
   */
  public void setRadius(double r) {
    this.r = r;
    trigger();
  }
  
  /**
   * Gets the height of the mirror, measured from the optical axis.
   * The extension of the mirror (i.e. the cross section of its projection onto
   * the tangential plane) is twice the height.
   */
  public double getHeight() {
    return this.h;
  }
  
  /**
   * Sets the height of the mirror, measured from the optical axis.
   * The extension of the mirror (i.e. the cross section of its projection onto
   * the tangential plane) is twice the height.
   */
  public void setHeight(double h) {
    if (h <= Math.abs(r)) {
      this.h = h;
      trigger();
    }
  }
  
  /**
   * Returns the position of the focal point as a one element Point2dVector.
   */
  public Point2dVector getFocalPoints() {
    Point2dVector point = new Point2dVector();
    point.addPoint(pos + r/2, 0.0);
    return point;
  }
  
  /**
   * Computes the image of the object src.<br>
   * Its rays connect the image point with the mirror and are either part of the
   * source rays (for real images) or are backwards extensions of the source
   * rays (for virtual images).
   */
  public LightSource getImage(ObjectLightSource src) {
    GenericLightSource image = new GenericLightSource();
    LightRayBundle rays = new LightRayBundle();
    
    DVector srcPos = src.getCoords();
    DVector imgPos = getImagePoint(srcPos);
    image.setCoords(imgPos);
    
    // compute rays from the image to the mirror
    int nRays = src.getRays().getNRays();
    DVector angles = src.getStartAngles();
    for (int i = 0; i < nRays; i++) {
      LightRay ray = new LightRay(imgPos, Double.NaN);
      double ys = getHittingHeight(srcPos, angles.get(i));
      if (Math.abs(ys) <= h) {
        ray.addPoint(pos, ys);
      }
      rays.addRay(ray);
    }
    
    image.setRays(rays);
    return image;
  }
  
  /**
   * Computes the position of the image of an object at position (xSrc, ySrc).
   */
  protected DVector getImagePoint(DVector p) {
    double a = pos - p.x();
    double f = r/2;
    double ap = -1/(1/f + 1/a);
    return new DVector(pos - ap, -p.y()*ap/a);
  }
  
  /**
   * Computes the height at which a ray, given by a start point and an angle,
   * hits the principal plane.
   */
  protected double getHittingHeight(DVector p, double angle) {
    return p.y() + (pos - p.x())*Math.tan(angle);
  }
  
  /**
   * Returns the smallest x value to which the mirror extends
   */
  protected double getLeftEdge() {
    return (r > 0)? pos : pos + r + Math.sqrt(r*r - h*h);
  }
  
  /**
   * Returns the largest x value to which the mirror extends
   */
  protected double getRightEdge() {
    return (r > 0)? pos + r - Math.sqrt(r*r - h*h) : pos;
  }
  
  /**
   * Returns the reflection angle of an incoming ray using the paraxial
   * approximation. The incoming ray is specified by the start point and
   * the hit height at the principal plane.
   */
  protected double getApproximateReflectionAngle(DVector pStart, double yHit) {
    DVector pImg = getImagePoint(pStart);
    double beta;
    if (Double.isInfinite(pImg.x())) {
      beta = Math.atan(pStart.y()/(pos - pStart.x()));
    } else {
      beta = Math.atan((yHit - pImg.y())/(pos - pImg.x()));
    }
    
    if (pStart.x() < pos) {
      beta -= Math.PI;
    }
    return beta;
  }
  
  /**
   * Gets the flag for paraxial aproximation.
   * If true rays are propagated by using the principal plane and the ideal
   * image construction. Otherwise the correct reflection points and angles
   * are computed.
   */
  public boolean isParaxialApproximation()  {
    return this.paraxialApproximation;
  }
  
  /**
   * Sets the flag for paraxial aproximation.
   * If true rays are propagated by using the principal plane and the ideal
   * image construction. Otherwise the correct reflection points and angles
   * are computed.
   */
  public void setParaxialApproximation(boolean paraxialApproximation)  {
    this.paraxialApproximation = paraxialApproximation;
    trigger();
  }
}
