package physbeans.phys;

import java.io.*;
import physbeans.model.*;
import physbeans.math.*;

/**
 * A lens using the usual paraxial and thin lens approximations.
 */
public class ThinLens extends Lens implements Serializable {
  
  /** Creates a new instance of ThinLens */
  public ThinLens() {
  }
  
  /**
   * Gets focal length of lens.
   * Uses simple lensmaker equation for a thin lens
   */
  public double getFocalLength() {
    return 1.0/((n - 1)*(1.0/r1 - 1.0/r2));
  }
  
  /**
   * Sets focal length of lens.
   * Sets index of refraction using the simple lensmaker equation for a thin
   * lens to get the requested f. According to the sign of f the signs of the
   * curvature radii are adapted to make the lens convex or concave.
   */
  public void setFocalLength(double f) {
    double nNew = 1.0/((1.0/r1 - 1.0/r2)*f) + 1;
    if (Maths.isStrictlyPositive(1.0 - nNew)) {
      r1 = -r1;
      r2 = -r2;
      nNew = 1.0/((1.0/r1 - 1.0/r2)*f) + 1;
    }
    n = nNew;
    trigger();
  }
  
  /**
   * Returns the position of the focal points as a Point2dVector.
   */
  public Point2dVector getFocalPoints() {
    Point2dVector point = new Point2dVector();
    double fp = getFocalLength();
    point.addPoint(pos - fp, 0.0);
    point.addPoint(pos + fp, 0.0);
    return point;
  }
  
  /**
   * Computes the image of the object src.<br>
   * Its rays connect the image point with the lens and are either part of the
   * source rays (for real images) or are backwards extensions of the source
   * rays (for virtual images).
   */
  public LightSource getImage(ObjectLightSource src) {
    GenericLightSource image = new GenericLightSource();
    LightRayBundle rays = new LightRayBundle();
    
    DVector srcPos = src.getCoords();
    DVector imgPos = getImagePoint(srcPos);
    image.setCoords(imgPos);
    
    // compute rays from the image to the mirror
    int nRays = src.getRays().getNRays();
    DVector angles = src.getStartAngles();
    for (int i = 0; i < nRays; i++) {
      LightRay ray = new LightRay(imgPos, Double.NaN);
      double ys = getHittingHeight(srcPos, angles.get(i));
      if (Math.abs(ys) <= getHeight()) {
        ray.addPoint(pos, ys);
      }
      rays.addRay(ray);
    }
    
    image.setRays(rays);
    return image;
  }
  
  /**
   * Computes the position of the image of an object at position (xSrc, ySrc).
   */
  protected DVector getImagePoint(DVector p) {
    double a = p.x() - pos;
    double fp = getFocalLength();
    fp = (a>0)? -fp: fp;
    double ap = 1.0/(1.0/fp + 1.0/a);
    return new DVector(pos + ap, p.y()*ap/a);
  }
  
  /**
   * Computes the height at which a ray, given by a start point and an angle,
   * hits the principal plane.
   */
  protected double getHittingHeight(DVector p, double angle) {
    return p.y() + (pos - p.x())*Math.tan(angle);
  }
  
  /**
   * Returns the refraction angle of an incoming ray using the paraxial
   * approximation. The incoming ray is specified by the start point and
   * the hit height at the principal plane.
   */
  protected double getApproximateRefractionAngle(DVector pStart, double yHit) {
    DVector pImg = getImagePoint(pStart);
    double beta;
    if (Double.isInfinite(pImg.x())) {
      beta = Math.atan(pStart.y()/(pStart.x() - pos));
    } else {
      beta = Math.atan((yHit - pImg.y())/(pos - pImg.x()));
    }
    
    if (pStart.x() > pos) {
      beta -= Math.PI;
    }
    return beta;
  }
  
  /**
   * Extends a light ray through the optical element.
   */
  protected void extend(LightRay ray) {
    if (rayIsDeparting(ray)) {
      return;
    }
    
    DVector pLast = ray.getLastPoint();
    double al = ray.getAngle();
    
    double y1 = getHittingHeight(pLast, al);
    if (Math.abs(y1) > getHeight()) {
      return;         // miss
    }
    ray.addPoint(pos, y1);
    ray.setAngle(getApproximateRefractionAngle(pLast, y1));
  }
  
  /**
   * Returns the position of the first principal plane.
   * Overwrites Lens and simply returns the lens midpoint position.
   */
  public double getFirstPrincipalPlane() {
    return pos;
  }
  
  /**
   * Returns the position of the second principal plane.
   * Overwrites Lens and simply returns the lens midpoint position.
   */
  public double getSecondPrincipalPlane() {
    return pos;
  }
}
