package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;
import physbeans.views.*;
import physbeans.func.*;

/**
 * Shows a cylinder with a moving piston.<br>
 * Optionally an additional quantity - usually the temperature - can be
 * displayed with a color code.
 */
public class CylinderView extends RestartableView implements Serializable {
  
  protected double lCyl, dCyl;
  protected double vMax;
  protected double lPiston, dPiston;
  
  protected ScalingFunction scaler;
  protected ColorTable colorTable;
  
  /**
   * Creates a standard PendulumView.
   */
  public CylinderView() {
    super();
    
    lCyl = 0.8;
    dCyl = 0.6;
    lPiston = 1.0;
    dPiston = 0.05;
    vMax = lCyl*Math.PI*dCyl*dCyl/4.0;
    
    scaler = new ScalingFunction(); // scales T values to color indices
    scaler.setXMin(200);
    scaler.setXMax(3000);
    scaler.setYMin(0);
    scaler.setYMax(255);
    scaler.setLogarithmic(true);
    colorTable = new HeatColorTable();
    
    setInitialValues(new DVector(0.5, 300.0));
  }
  
  /**
   * Redraws the complete cylinder and piston according to driving parameters.
   */
  public void paintLayer(Graphics2D g) {
    double x = model.get(0)*lCyl/vMax;
    double T = model.get(1);
    
    drawCylinder(g);
    drawPiston(g, x);
    drawGas(g,x, T);
  }
  
  /**
   * Sets volume of cylinder.
   * Conveniance function to simplify routing.
   */
  public void setVolume(double v) {
    model.set(0, v);
    update();
  }
  
  /**
   * Sets second parameter (usually temperature inside cylinder).
   * Conveniance function to simplify routing.
   */
  public void setTemperature(double t) {
    model.set(1, t);
    update();
  }
  
  /**
   * Gets maximal volume of the cylinder.
   */
  public double getMaxVolume() {
    return this.vMax;
  }
  
  /**
   * Sets maximal volume of the cylinder.
   */
  public void setMaxVolume(double vMax) {
    this.vMax = vMax;
    update();
  }
  
  /**
   * Gets length of the cylinder.
   */
  public double getLength() {
    return this.lCyl;
  }
  
  /**
   * Sets length of the cylinder.
   */
  public void setLength(double length) {
    this.lCyl = lCyl;
    update();
  }
  
  /**
   * Gets minimal value for temperature scale.
   */
  public double getTMin() {
    return scaler.getXMin();
  }
  
  /**
   * Sets minimal value for temperature scale.
   */
  public void setTMin(double tMin) {
    scaler.setXMin(tMin);
    update();
  }
  
  /**
   * Gets maximal value for temperature scale.
   */
  public double getTMax() {
    return scaler.getXMax();
  }
  
  /**
   * Sets maximal value for temperature scale.
   */
  public void setTMax(double tMax) {
    scaler.setXMax(tMax);
    update();
  }
  
  /**
   * Gets diameter of the cylinder.
   * It is only used for drawing purposes and not related to the maximal volume
   * of the cylinder.
   */
  public double getDiameter() {
    return this.dCyl;
  }
  
  /**
   * Sets diameter of the cylinder.
   * It is only used for drawing purposes and not related to the maximal volume
   * of the cylinder.
   */
  public void setDiameter(double diameter) {
    this.dCyl = dCyl;
    update();
  }
  
  /**
   * Gets length of the piston.
   */
  public double getPistonLength() {
    return this.lPiston;
  }
  
  /**
   * Sets length of the piston.
   */
  public void setPistonLength(double lPiston) {
    this.lPiston = lPiston;
    update();
  }
  
  /**
   * Gets diameter of the piston.
   */
  public double getPistonDiameter() {
    return this.dPiston;
  }
  
  /**
   * Sets diameter of the piston.
   */
  public void setPistonDiameter(double dPiston) {
    this.dPiston = dPiston;
    update();
  }
  
  /**
   * Draws the cylinder wall.
   */
  protected void drawCylinder(Graphics2D g) {
    // the center of the ground plate of the cylinder is at (0,0)
    float s1 =(float) lCyl;
    float s2 = 0.0f;
    float t1 = (float) dCyl/2.0f;
    float t2 = (float) -dCyl/2.0f;
    
    GeneralPath path = new GeneralPath();
    path.moveTo(s1, t1);
    path.lineTo(s2, t1);
    path.lineTo(s2, t2);
    path.lineTo(s1, t2);
    g.draw(path);
  }
  
  /**
   * Draws the piston of the cylinder ending at position x.
   */
  protected void drawPiston(Graphics2D g, double x) {
    Line2D line = new Line2D.Double(x, -dCyl/2.0, x, dCyl/2.0);
    Rectangle2D rect = new Rectangle2D.Double(x, -dPiston/2.0, lPiston, dPiston);
    g.draw(line);
    g.fill(rect);
  }
  
  /**
   * Draws the gas inside the cylinder, colored according to its temperature.
   */
  protected void drawGas(Graphics2D g, double x, double T) {
    Rectangle2D cyl = new Rectangle2D.Double();
    cyl.setRect(0.0, -dCyl/2.0, x, dCyl);
    
    int col = (int) Math.round(scaler.evaluate(T));
    g.setColor(colorTable.getColor(col));
    g.fill(cyl);
    g.setColor(Color.BLACK);
    g.draw(cyl);
  }
}
