package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;

/**
 * Displays a set of labels horizontally at equal distances (i.e. on a
 * grid).<br>
 * The labels can be simple strings or they can be set by a vector of numerical
 * values.
 */
public class LabelView extends View implements Serializable {

  protected String[] model;
  protected String[] initialValues;
  protected int precision;
  protected Color backgroundColor;
  protected Color textColor;
  protected double position;
  protected double marginLeft;
  protected double marginRight;
  protected double padding;

  /**
   * Creates a standard LabelView.
   */
  public LabelView() {
    super();

    model = new String[]{"1.0", "2.0", "3.0"};
    initialValues = new String[]{"1.0", "2.0", "3.0"};
    precision = 2;
    backgroundColor = null;
    textColor = Color.BLACK;
    position = 0.0;    // at the top
    marginLeft = 0.1;
    marginRight = 0.1;
    padding = 0.01;
  }

  public void paintLayer(Graphics2D g) {
    // drawing strings is done best in pixels
    double xScale = trafo.getScalingX();
    int marginLeftPix = (int) (marginLeft / xScale);
    int marginRightPix = (int) (marginRight / xScale);
    int paddingPix = (int) (padding / xScale);
    AffineTransform oldTrafo = g.getTransform();
    g.transform(trafo.getInverseAffineTransform());

    int width = trafo.getMaxS();

    int distance = (width - marginLeftPix - marginRightPix) / model.length;
    FontMetrics metrics = g.getFontMetrics();
    int h = metrics.getAscent() + metrics.getDescent();   // height of the labels
    int hTotal = trafo.getMaxT() - h - 2 * paddingPix;
    int vPos = (int) Math.round(h / 2.0 + paddingPix + position * hTotal);
    // vertical position of label centers
    int dh = h / 2 - metrics.getDescent();        // from midpoint to baseline

    int hPos = marginLeftPix + distance / 2;         // position of next label center
    int w = 0;                                  // width of a label
    for (int i = 0; i < model.length; i++) {    // draw labels
      try {
        w = metrics.stringWidth(model[i]);
        // draw background rectangle
        if (backgroundColor != null) {
          g.setColor(backgroundColor);
          g.fillRect(hPos - w / 2 - paddingPix, vPos - h / 2 - paddingPix,
                  w + 2 * paddingPix, h + 2 * paddingPix);
        }
        // draw label
        g.setColor(textColor);
        g.drawString(model[i], hPos - w / 2, vPos + h / 2 - dh);
      } catch (NullPointerException e) {
        // sometimes model[i] == null due to thread-problems
        // just ignore it
      }

      //proceed
      hPos += distance;
    }

    g.setTransform(oldTrafo);
  }

  /**
   * Gets the displayed string given by the index.
   *
   * @return Value of the property at <CODE>index</CODE>.
   */
  public String getModel(int index) {
    return this.model[index];
  }

  /**
   * Gets the model of the view, here the array of strings.
   *
   * @return Value of property model.
   */
  public String[] getModel() {
    return this.model;
  }

  /**
   * Sets the displayed string given by the index.
   *
   * @param index Index of the property.
   * @param model New value of the property at <CODE>index</CODE>.
   */
  public void setModel(int index, String model) {
    this.model[index] = model;
    update();
  }

  /**
   * Sets the model of the view, here the array of strings.
   *
   * @param model New value of property model.
   */
  public void setModel(String[] model) {
    this.model = model;
    update();
  }

  /**
   * Sets the labels using the values of a DVector.<br>
   * For the display of the numbers, the defined precision is used.
   */
  public void setValues(DVector values) {
    int n = values.getDimension();
    model = new String[n];

    for (int i = 0; i < n; i++) {
      model[i] = Maths.toString(values.get(i), precision);
    }

    update();
  }

  /**
   * Sets the first label.<br>
   * Conveniance function for a 1d LabelView.
   */
  public void setValue(double value) {
    setValues(new DVector(value));
  }

  /**
   * Gets the number of decimals to be displayed, if the labels are defined by a
   * DVector.
   *
   * @return Value of property precision.
   */
  public int getPrecision() {
    return this.precision;
  }

  /**
   * Sets the number of decimals to be displayed, if the labels are defined by a
   * DVector.
   *
   * @param precision New value of property precision.
   */
  public void setPrecision(int precision) {
    this.precision = precision;
    update();
  }

  /**
   * Gets the background color used for the lables.<br>
   * If it is null, the background color of the parent is used.
   *
   * @return Value of property backgroundColor.
   */
  public Color getBackgroundColor() {
    return this.backgroundColor;
  }

  /**
   * Sets the background color used for the lables.<br>
   * If it is null, the background color of the parent is used.
   *
   * @param backgroundColor New value of property backgroundColor.
   */
  public void setBackgroundColor(Color backgroundColor) {
    this.backgroundColor = backgroundColor;
    update();
  }

  /**
   * Gets the text color used for the lables.
   *
   * @return Value of property textColor.
   */
  public Color getTextColor() {
    return this.textColor;
  }

  /**
   * Sets the text color used for the lables.
   *
   * @param textColor New value of property textColor.
   */
  public void setTextColor(Color textColor) {
    this.textColor = textColor;
    update();
  }

  /**
   * Gets the (vertical) position of the labels relative to the panel size.<br>
   * position 0.0 means at the top, 1.0 at the bottom (accounting for the height
   * of the LabelView).
   *
   * @return Value of property position.
   */
  public double getPosition() {
    return this.position;
  }

  /**
   * Gets the (vertical) position of the labels relative to the panel size.<br>
   * position 0.0 means at the top, 1.0 at the bottom (accounting for the height
   * of the LabelView).
   *
   * @param position New value of property position.
   */
  public void setPosition(double position) {
    this.position = position;
    update();
  }

  /**
   * Gets the size of the left margin of the label view.
   *
   * @return Value of property marginLeft.
   */
  public double getMarginLeft() {
    return this.marginLeft;
  }

  /**
   * Sets the size of the left margin of the label view.
   *
   * @param marginLeft New value of property marginLeft.
   */
  public void setMarginLeft(double marginLeft) {
    this.marginLeft = marginLeft;
    update();
  }

  /**
   * Gets the size of the right margin of the label view.
   *
   * @return Value of property marginRight.
   */
  public double getMarginRight() {
    return this.marginRight;
  }

  /**
   * Sets the size of the right margin of the label view.
   *
   * @param marginRight New value of property marginRight.
   */
  public void setMarginRight(double marginRight) {
    this.marginRight = marginRight;
    update();
  }

  /**
   * Gets the size of the padding surrounding one label.
   *
   * @return Value of property padding.
   */
  public double getPadding() {
    return this.padding;
  }

  /**
   * Sets the size of the padding surrounding one label.
   *
   * @param padding New value of property padding.
   */
  public void setPadding(double padding) {
    this.padding = padding;
    update();
  }

  /**
   * Gets initial values of all model parameters.
   */
  public String[] getInitialValues() {
    return initialValues;
  }

  /**
   * Sets all initial values.<br>
   * Sets the actual values to the initial values.
   */
  public void setInitialValues(String[] vals) {
    initialValues = vals;
    model = new String[vals.length];
    System.arraycopy(vals, 0, model, 0, vals.length);
    update();
  }

  /**
   * Sets all initial values from a DVector.
   */
  public void setInitialValues(DVector values) {
    int n = values.getDimension();
    String[] sVals = new String[n];

    for (int i = 0; i < n; i++) {
      sVals[i] = Maths.toString(values.get(i), precision);
    }

    setInitialValues(sVals);
  }

  /**
   * Resets the model to the initial values and redraws.
   */
  public void restart() {
    System.arraycopy(initialValues, 0, model, 0, initialValues.length);
    update();
  }
}
