package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;

import physbeans.inout.ScreenWorldTrafo;
import physbeans.phys.*;

/**
 * Displays a spherical lens.<p>
 * Additionally the active plane and the focal points can be shown.
 */
public class LensView extends View implements Serializable {
  
  protected Lens model;
  protected Color lineColor;
  protected Color lensColor;
  protected Color principalPlaneColor;
  protected boolean showPrincipalPlane;
  protected boolean showFocalPoints;
  protected PointView focalView;
  
  /**
   * creates a LensView with default values
   */
  public LensView() {
    model = new Lens();
    lineColor = Color.black;
    lensColor = Color.white;
    principalPlaneColor = new Color(128, 128, 255);
    showPrincipalPlane = true;
    showFocalPoints = false;
    focalView = new PointView();
    focalView.setPointRadius(0.2);
    focalView.setPointColor(Color.black);
  }
  
  public void paintLayer(Graphics2D g2) {
    if (model == null) {
      return;  // do nothing
    }
    
    double x0 = model.getPosition();
    double x1 = model.getCurvatureCenter1();
    double r1 = model.getRadius1();
    double x2 = model.getCurvatureCenter2();
    double r2 = model.getRadius2();
    double h = model.getHeight();
    
    Area lens;
    if (model.isConvex()) {
      // construct lens as intersection of two circles
      Ellipse2D.Double c1 = new Ellipse2D.Double();
      c1.setFrameFromCenter(x1, 0.0, x1 + r1, r1);
      Ellipse2D.Double c2 = new Ellipse2D.Double();
      c2.setFrameFromCenter(x2, 0.0, x2 + r2, r2);
      lens = new Area(c1);
      lens.intersect(new Area(c2));
      
      // constrain height
      Rectangle2D.Double rect = new Rectangle2D.Double();
      rect.setFrameFromCenter(x0, 0.0, Math.max(2*r1, -2*r2) + x0, h);
      lens.intersect(new Area(rect));
    } else {
      // assuming concave
      double d = model.getThickness();
      double b1 = model.getLeftEdge1();
      double b2 = model.getRightEdge2();
      
      // construct lens as rectangle - two circles
      Rectangle2D.Double rect = new Rectangle2D.Double();
      rect.setFrameFromDiagonal(b1, -h, b2, h);
      Ellipse2D.Double c1 = new Ellipse2D.Double();
      c1.setFrameFromCenter(x1, 0.0, x1 + r1, r1);
      Ellipse2D.Double c2 = new Ellipse2D.Double();
      c2.setFrameFromCenter(x2, 0.0, x2 + r2, r2);
      lens = new Area(rect);
      lens.subtract(new Area(c1));
      lens.subtract(new Area(c2));
    }
    
    g2.setColor(lensColor);
    g2.fill(lens);
    if (showPrincipalPlane) {
      g2.setColor(principalPlaneColor);
      double xh1 = model.getFirstPrincipalPlane();
      double xh2 = model.getSecondPrincipalPlane();
      Line2D line1 = new Line2D.Double(xh1, h, xh1, -h);
      Line2D line2 = new Line2D.Double(xh2, h, xh2, -h);
      g2.draw(line1);
      g2.draw(line2);
    }
    g2.setColor(lineColor);
    g2.draw(lens);
    
    if (showFocalPoints) {
      focalView.setModel(model.getFocalPoints());
      focalView.paintLayer(g2);
    }
  }
  
  /**
   * Gets lens to paint.
   */
  public Lens getModel() {
    return this.model;
  }
  
  /**
   * Sets lens to paint.
   */
  public void setModel(Lens model) {
    this.model = model;
    update();
  }
  
  /**
   * Gets color of the lens boundary.
   */
  public Color getLineColor() {
    return this.lineColor;
  }
  
  /**
   * Sets color of the lens boundary.
   */
  public void setLineColor(Color lineColor) {
    this.lineColor = lineColor;
    update();
  }
  
  /**
   * Gets color of the lens.
   */
  public Color getLensColor() {
    return this.lensColor;
  }
  
  /**
   * Sets color of the lens.
   */
  public void setLensColor(Color lensColor) {
    this.lensColor = lensColor;
    update();
  }
  
  /**
   * Gets the line color of the active plane.
   */
  public Color getPrincipalPlaneColor() {
    return this.principalPlaneColor;
  }
  
  /**
   * Sets the line color of the active plane.
   */
  public void setPrincipalPlaneColor(Color principalPlaneColor) {
    this.principalPlaneColor = principalPlaneColor;
    update();
  }
  
  /**
   * Gets the value of the showFocalPoints flag.
   */
  public boolean isShowFocalPoints() {
    return this.showFocalPoints;
  }
  
  /**
   * Sets the value of the showFocalPoints flag.
   */
  public void setShowFocalPoints(boolean showFocalPoints) {
    this.showFocalPoints = showFocalPoints;
    update();
  }
  
  /**
   * Gets the value of the showPrincipalPlane flag.
   */
  public boolean isShowPrincipalPlane() {
    return this.showPrincipalPlane;
  }
  
  /**
   * Sets the value of the showPrincipalPlane flag.
   */
  public void setShowPrincipalPlane(boolean showPrincipalPlane) {
    this.showPrincipalPlane = showPrincipalPlane;
    update();
  }
  
  /**
   * Gets radius of the circle representing the focal point.
   */
  public double getFocalPointRadius() {
    return focalView.getPointRadius();
  }
  
  /**
   * Sets radius of the circle representing the focal point.
   */
  public void setFocalPointRadius(double r) {
    focalView.setPointRadius(r);
    update();
  }
  
  /**
   * Sets trafo between screen and world coordinates.
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    super.setTrafo(trafo);
    focalView.setTrafo(trafo);
  }
  
}
