package physbeans.views;

import java.awt.*;
import java.io.*;
import physbeans.inout.ScreenWorldTrafo;
import physbeans.math.*;

/**
 * Displays a swinging horizontal/vertical spring with attached mass.
 * Optionally it display an additional force vector.
 */
public class SpringView extends RestartableView implements Serializable {

  protected boolean showArrow;
  protected SingleSpringView spring;
  protected SingleMassView mass;
  protected SingleVectorView force;
  protected DVector mPos;

  /**
   * create a standard horizontal SpringView
   */
  public SpringView() {
    super();

    showArrow = false;
    model = new DVector(0.5, 0.5);
    initialValues = new DVector(0.5, 0.5);

    spring = new SingleSpringView();
    mass = new SingleMassView();
    force = new SingleVectorView();

    spring.setEndLineLength(spring.getEndLineLength() + mass.getMassWidth() / 2);

    mPos = new DVector(2);
  }

  /**
   * redraw the complete spring according to driving parameter
   */
  public void paintLayer(Graphics2D g) {
    double x = model.get(0);

    mPos.set(0, x);

    spring.setRightPosition(mPos);
    spring.paintLayer(g);
    mass.setPosition(mPos);
    mass.paintLayer(g);

    if (showArrow) {
      force.setPosition(mPos);
      force.setValue(model.get(1));
      force.paintLayer(g);
    }
  }

  /**
   * Gets width of the mass.
   */
  public double getMassWidth() {
    return mass.getMassWidth();
  }

  /**
   * Sets width of the mass.
   */
  public void setMassWidth(double mWidth) {
    mass.setMassWidth(mWidth);
  }

  /**
   * Gets height of the mass.
   */
  public double getMassHeight() {
    return mass.getMassHeight();
  }

  /**
   * Sets height of the mass.
   */
  public void setMassHeight(double mHeight) {
    mass.setMassHeight(mHeight);
  }

  /**
   * Gets number of turns of the spring.
   */
  public int getNTurns() {
    return spring.getNTurns();
  }

  /**
   * Sets number of turns of the spring.
   */
  public void setNTurns(int nTurns) {
    spring.setNTurns(nTurns);
  }

  /**
   * Gets width of the spring.
   */
  public double getSpringWidth() {
    return spring.getSpringWidth();
  }

  /**
   * Sets width of the spring.
   */
  public void setSpringWidth(double springWidth) {
    spring.setSpringWidth(springWidth);
  }

  /**
   * Gets position of the point where the spring is attached to.
   */
  public double getFixPosition() {
    return spring.getLeftPosition().x();
  }

  /**
   * Sets position of the point where the spring is attached to.
   */
  public void setFixPosition(double fixPosition) {
    spring.setLeftPosition(new DVector(fixPosition, 0));
  }

  /**
   * Gets length of the straight line at the beginning of the spring
   */
  public double getStartLineLength() {
    return spring.getStartLineLength();
  }

  /**
   * Sets length of the straight line at the beginning of the spring
   */
  public void setStartLineLength(double lBegin) {
    spring.setStartLineLength(lBegin);
  }

  /**
   * Gets length of the straight line at the end of the spring
   */
  public double getEndLineLength() {
    return spring.getEndLineLength();
  }

  /**
   * Sets length of the straight line at the end of the spring
   */
  public void setEndLineLength(double lEnd) {
    spring.setEndLineLength(lEnd);
  }

  /**
   * gets the color of the force arrow
   */
  public Color getArrowColor() {
    return force.getArrowColor();
  }

  /**
   * sets the color of the force arrow
   */
  public void setArrowColor(Color arrowColor) {
    force.setArrowColor(arrowColor);
  }

  /**
   * Getser the arrowScale.
   * @return Value of property arrowScale.
   */
  public double getArrowScale() {
    return force.getArrowScale();
  }

  /**
   * Sets the arrowScale.
   * @param arrowScale New value of property arrowScale.
   */
  public void setArrowScale(double arrowScale) {
    force.setArrowScale(arrowScale);
  }

  /**
   * gets length of arrow tip relative to arrow length
   */
  public double getArrowTipLength() {
    return force.getArrowTipLength();
  }

  /**
   * sets length of arrow tip relative to arrow length
   */
  public void setArrowTipLength(double arrowTipLength) {
    force.setArrowTipLength(arrowTipLength);
  }

  /**
   * gets width of arrow tip relative to arrow length
   */
  public double getArrowTipWidth() {
    return force.getArrowTipWidth();
  }

  /**
   * sets width of arrow tip relative to arrow length
   */
  public void setArrowTipWidth(double arrowTipWidth) {
    force.setArrowTipWidth(arrowTipWidth);
  }

  /**
   * Getter for property showArrow.
   * @return Value of property showArrow.
   */
  public boolean isShowArrow() {
    return this.showArrow;
  }

  /**
   * Setter for property showArrow.
   * @param showArrow New value of property showArrow.
   */
  public void setShowArrow(boolean showArrow) {
    this.showArrow = showArrow;
    update();
  }

  /**
   * used by LayeredScreen
   */
  public void setTrafo(ScreenWorldTrafo trafo) {
    super.setTrafo(trafo);
    spring.setTrafo(trafo);
    mass.setTrafo(trafo);
    force.setTrafo(trafo);
  }
}
