package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;
import physbeans.func.*;

/**
 * Shows a schematic diagram of a gas turbine consisting of a compressor,
 * a heater, a turbine and a cooler.<br>
 * The temperature of the gas in one of the four parts (the "active" part) is
 * displayed with a color code. An additional input defines the number of this
 * part.
 */
public class TurbineView extends View implements Serializable {
  
  protected DVector model;
  protected ScalingFunction scaler;
  protected ColorTable colorTable;
  
  /**
   * Creates a standard TurbineView.
   */
  public TurbineView() {
    super();
    model = new DVector(300.0, 0.0);
    
    scaler = new ScalingFunction(); // scales T values to color indices
    scaler.setXMin(200);
    scaler.setXMax(3000);
    scaler.setYMin(0);
    scaler.setYMax(255);
    scaler.setLogarithmic(true);
    colorTable = new HeatColorTable();
  }
  
  /**
   * Redraws the complete turbine according to driving parameters.
   */
  public void paintLayer(Graphics2D g) {
    int phase = (int) Math.round(model.get(1));
    
    drawCompressorV(g, new DVector(-1.0,0.0), 0.6, (phase%4 == 0));
    drawHeaterH(g, new DVector(0.0,1.0), 0.8, (phase%4 == 1));
    drawTurbineV(g, new DVector(1.0,0.0), 0.55, (phase%4 == 2));
    drawHeaterH(g, new DVector(0.0,-1.0), 0.8, (phase%4 == 3));
  }
  
  /**
   * Gets underlying model defining the temperature and the phase.
   */
  public DVector getModel() {
    return model;
  }
  
  /**
   * Sets underlying model defining the temperature and the phase.
   */
  public void setModel(DVector model) {
    this.model = model;
    update();
  }
  
  /**
   * Sets phase of the turbine process. This defines the part which is colored
   * according to its temperature.
   * Conveniance function to simplify routing.
   */
  public void setPhase(double v) {
    model.set(1, v);
    update();
  }
  
  /**
   * Sets temperature of the gas.
   * Conveniance function to simplify routing.
   */
  public void setTemperature(double t) {
    model.set(0, t);
    update();
  }
  
  /**
   * Gets minimal value for temperature scale.
   */
  public double getTMin() {
    return scaler.getXMin();
  }
  
  /**
   * Sets minimal value for temperature scale.
   */
  public void setTMin(double tMin) {
    scaler.setXMin(tMin);
    update();
  }
  
  /**
   * Gets maximal value for temperature scale.
   */
  public double getTMax() {
    return scaler.getXMax();
  }
  
  /**
   * Sets maximal value for temperature scale.
   */
  public void setTMax(double tMax) {
    scaler.setXMax(tMax);
    update();
  }
  
  /**
   * Draws a compressor symbol at position (-1,0) with connecting pipes of
   * length l.
   * If it is active is shows the temperature using the color table.
   */
  protected void drawCompressorV(Graphics2D g, DVector v, double l, boolean active) {
    g.translate(v.x(), v.y());
    
    double r = 0.4;
    double phi1 = 10.0*Math.PI/180.0;
    double phi2 = 30.0*Math.PI/180.0;
    
    Point2D p0a = new Point2D.Double(- r, r);
    Point2D p0c = new Point2D.Double(0.0, 0.0);
    Point2D p1 = new Point2D.Double(r*Math.sin(phi1), r*Math.cos(phi1));
    Point2D p2 = new Point2D.Double(r*Math.sin(phi2), -r*Math.cos(phi2));
    Point2D p3 = new Point2D.Double(-r*Math.sin(phi1), r*Math.cos(phi1));
    Point2D p4 = new Point2D.Double(-r*Math.sin(phi2), -r*Math.cos(phi2));
    
    Ellipse2D.Double c = new Ellipse2D.Double();
    c.setFrameFromCenter(p0c, p0a);
    
    // workaround for bug in Java 1.4: fill(c) doesn't fill draw(c) completely!
    double eps = 0.008;
    Point2D p0b = new Point2D.Double(- r - eps, r + eps);
    Ellipse2D.Double cf = new Ellipse2D.Double();
    cf.setFrameFromCenter(p0c, p0b);
    
    if (active) {
      g.setColor(getColor());
      g.fill(cf);
      g.setColor(Color.BLACK);
    }
    g.draw(c);
    
    Line2D.Double l1 = new Line2D.Double(p1, p2);
    Line2D.Double l2 = new Line2D.Double(p3, p4);
    g.draw(l1);
    g.draw(l2);
    
    // connecting pipes
    Line2D.Double la = new Line2D.Double(0.0, r+l, 0.0, r);
    Line2D.Double lb = new Line2D.Double(0.0, -r-l, 0.0, -r);
    g.draw(la);
    g.draw(lb);
    
    g.translate(-v.x(), -v.y());
  }
  
  /**
   * Draws a vertical turbine symbol at position (1,0) with connecting pipes of
   * length l.
   * If it is active is shows the temperature using the color table.
   */
  protected void drawTurbineV(Graphics2D g, DVector v, double l, boolean active) {
    g.translate(v.x(), v.y());
    
    double b = 0.4;
    double h1 = 0.6;
    double h2 = 1.2;
    
    Point2D p0 = new Point2D.Double(0.0, 0.0);
    Point2D p1 = new Point2D.Double(- b/2, -h1/2);
    Point2D p2 = new Point2D.Double(- b/2, h1/2);
    Point2D p3 = new Point2D.Double(b/2, h2/2);
    Point2D p4 = new Point2D.Double(b/2, -h2/2);
    
    GeneralPath poly = new GeneralPath();
    poly.moveTo((float) p1.getX(), (float) p1.getY());
    poly.lineTo((float) p2.getX(), (float) p2.getY());
    poly.lineTo((float) p3.getX(), (float) p3.getY());
    poly.lineTo((float) p4.getX(), (float) p4.getY());
    poly.closePath();
    
    if (active) {
      g.setColor(getColor());
      g.fill(poly);
      g.setColor(Color.BLACK);
    }
    g.draw(poly);
    
    // connecting pipes
    double mid = (h1 + h2)/4.0;
    Line2D.Double la = new Line2D.Double(0.0, mid + l, 0.0, mid);
    Line2D.Double lb = new Line2D.Double(0.0, -mid - l, 0.0, -mid);
    g.draw(la);
    g.draw(lb);
    
    g.translate(-v.x(), -v.y());
  }
  
  /**
   * Draws a horizontal heater/cooler symbol  with connecting pipes of length l
   * at position v.
   * If it is active is shows the temperature using the color table.
   */
  protected void drawHeaterH(Graphics2D g, DVector v, double l, boolean active) {
    g.translate(v.x(), v.y());
    
    double a = 0.4;
    Point2D pc = new Point2D.Double(0.0, 0.0);
    Point2D pe = new Point2D.Double(-a/2, -a/2);
    Point2D p1 = new Point2D.Double(-a/2, 0.0);
    Point2D p2 = new Point2D.Double(0.0, a/2);
    Point2D p3 = new Point2D.Double(a/2, 0.0);
    Point2D p4 = new Point2D.Double(0.0, -a/2);
    
    Rectangle2D.Double r = new Rectangle2D.Double();
    r.setFrameFromCenter(pc, pe);
    
    if (active) {
      g.setColor(getColor());
      g.fill(r);
      g.setColor(Color.BLACK);
    }
    g.draw(r);
    
    
    Line2D.Double l1 = new Line2D.Double(p1, p2);
    Line2D.Double l2 = new Line2D.Double(p2, p4);
    Line2D.Double l3 = new Line2D.Double(p4, p3);
    g.draw(l1);
    g.draw(l2);
    g.draw(l3);
    
    // connecting pipes
    Line2D.Double la = new Line2D.Double(-a/2, 0.0, -a/2-l, 0.0);
    Line2D.Double lb = new Line2D.Double(a/2, 0.0, a/2+l, 0.0);
    g.draw(la);
    g.draw(lb);
    
    g.translate(-v.x(), -v.y());
  }
  
  /**
   * Draws a vertical heater/cooler symbol  with connecting pipes of length l
   * at position v.
   * If it is active is shows the temperature using the color table.
   */
  protected void drawHeaterV(Graphics2D g, DVector v, double l, boolean active) {
    DVector nullVec = new DVector(2);
    
    g.translate(v.x(), v.y());
    g.rotate(Math.PI/2);
    drawHeaterH(g, nullVec, l, active);
    g.rotate(-Math.PI/2);
    g.translate(-v.x(), -v.y());
  }
  
  /**
   * Returns the color corresponding to the current temperature.
   */
  protected Color getColor() {
    int index = (int) Math.round(scaler.evaluate(model.get(0)));
    return colorTable.getColor(index);
  }
  
}
