package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import physbeans.math.*;

/**
 * Displays masses moving in 2d represented as spheres.
 */
public class TwoDMassesView extends RestartableView implements Serializable {

  protected DVector ri;    // mass radii
  protected double rM;     // marker radius
  protected Color color;   // sphere color
  protected boolean showLastAsMarker;

  /**
   * Create a standard square TwoDMassesView.
   */
  public TwoDMassesView() {
    super();
    model = new DVector(4);
    initialValues = new DVector(4);
    showLastAsMarker = false;
    ri = DVector.createFilledDVector(2, 0.4);
    rM = 0.1;
    color = Color.black;
  }

  /**
   * Draws the masses.
   */
  public void paintLayer(Graphics2D g) {
    // model = [x1, y1, x2, y2]
    int n = model.getDimension() / 2;
    int nMass = showLastAsMarker ? n - 1 : n;

    g.setColor(color);
    for (int i = 0; i < nMass; i++) {
      Arc2D.Double m = new Arc2D.Double();
      double r = ri.get(i);
      m.setArc(model.get(2 * i) - r, model.get(2 * i + 1) - r, 2 * r, 2 * r, 0, 360, 1);
      g.fill(m);
    }

    if (showLastAsMarker) {
      g.setColor(Color.black);
      drawMarker(g, model.get(2 * n - 2), model.get(2 * n - 1));
    }
  }

  /**
   * Gets width of the mass.
   */
  public DVector getRadii() {
    return this.ri;
  }

  /**
   * Sets width of the mass.
   */
  public void setRadii(DVector ri) {
    this.ri = ri;
    update();
  }

  /**
   * Gets position of the masses.
   */
  public DVector getPosition() {
    return getModel();
  }

  /**
   * Sets position of the masses.
   */
  public void setPosition(DVector position) {
    setModel(position);
  }

  /**
   * Gets value of showLastAsMarker flag.<br>
   * If true, the last position is shown as a simple marker. This is used for
   * denoting the CMS position.
   */
  public boolean isShowLastAsMarker() {
    return showLastAsMarker;
  }

  /**
   * Gets value of showLastAsMarker flag.<br>
   * If true, the last position is shown as a simple marker. This is used for
   * denoting the CMS position.
   */
  public void setShowLastAsMarker(boolean flag) {
    showLastAsMarker = flag;
  }

  /**
   * Sets the x-positions of the masses from a vector.<br>
   * Conveniance function for 1d systems.
   *
   * @param pos x coordinates of the masses
   */
  public void setPositions1D(DVector pos) {
    DVector position = getModel();
    int nMass = model.getDimension() / 2;
    for (int i = 0; i < nMass; i++) {
      position.set(2 * i, pos.get(i));
    }
    setPosition(position);
  }

  /**
   * Sets the initial x-positions from a vector.<br>
   * Conveniance function for 1d systems.
   *
   * @param pos x coordinates of the masses
   */
  public void setInitialValues1D(DVector pos) {
    DVector x0 = getInitialValues();
    int nMass = x0.getDimension() / 2;
    for (int i = 0; i < nMass; i++) {
      x0.set(2 * i, pos.get(i));
    }
    setInitialValues(x0);
  }

  /**
   * Gets color of spheres.
   */
  public Color getColor() {
    return this.color;
  }

  /**
   * Sets color of spheres.
   */
  public void setColor(Color color) {
    this.color = color;
    update();
  }

  /**
   * Gets radius of the circle representing the focal point.
   */
  public double getMarkerRadius() {
    return rM;
  }

  /**
   * Sets radius of the circle representing the focal point.
   */
  public void setMarkerRadius(double rM) {
    this.rM = rM;
  }

  protected void drawMarker(Graphics2D g, double x, double y) {
    // draws a marker at position (x,y)
    double dx = 0.70710678119 * rM;   // = cos(45°)
    Ellipse2D circle = new Ellipse2D.Double(x - rM, y - rM, 2 * rM, 2 * rM);
    g.draw(circle);
    Line2D l1 = new Line2D.Double(x - dx, y - dx, x + dx, y + dx);
    Line2D l2 = new Line2D.Double(x + dx, y - dx, x - dx, y + dx);
    g.draw(l1);
    g.draw(l2);
  }
}
