package physbeans.views;

import java.awt.*;
import java.awt.geom.*;
import physbeans.math.*;
import static java.lang.Math.*;

/**
 * Displays a vector at a given position together with its x- and y components.
 */
public class VectorDecompositionView extends RestartableView {
  // model = 4d vector posX/posY/vecX/vecY

  protected SingleVectorView v, vx, vy;
  protected boolean showMarker;
  protected int dotSize;
  protected boolean polar;

  /**
   * create a standard horizontal SpringView
   */
  public VectorDecompositionView() {
    super();

    model = new DVector(0.0, 0.0, 1.0, 1.0);
    initialValues = new DVector(0.0, 0.0, 1.0, 1.0);

    v = new SingleVectorView();
    vx = new SingleVectorView();
    vy = new SingleVectorView();

    showMarker = false;
    dotSize = 1;
    polar = false;

    setArrowColor(Color.red);
    setArrowTipWidth(0.1);
    setArrowTipLength(0.1);
    setArrowScale(0.5);
  }

  @Override
  public void paintLayer(Graphics2D g2) {
    // set position
    DVector pos = model.getDVector(0, 1);
    v.setPosition(pos);
    vx.setPosition(pos);
    vy.setPosition(pos);

    // set vector
    double x = model.get(2);
    double y = model.get(3);
    v.setValue(hypot(x, y));
    v.setAngle(atan2(y, x));

    if (polar) {
      double phi = atan2(pos.y(), pos.x());
      vx.setValue(x * cos(phi) + y * sin(phi));
      vx.setAngle(phi);
      vy.setValue(-x * sin(phi) + y * cos(phi));
      vy.setAngle(phi + PI / 2);
    } else {
      vx.setValue(abs(x));
      vx.setAngle(x > 0 ? 0 : PI);
      vy.setValue(abs(y));
      vy.setAngle(y > 0 ? PI / 2 : -PI / 2);
    }

    v.paintLayer(g2);
    vx.paintLayer(g2);
    vy.paintLayer(g2);

    if (showMarker) {
      drawPoint(g2, model.get(0), model.get(1));
    }
  }

  /**
   * Gets the vector.
   */
  public DVector getVector() {
    return model.getDVector(2, 3);
  }

  /**
   * Sets the vector.
   */
  public void setVector(DVector vector) {
    model.setDVector(2, 3, vector);
    update();
  }

  /**
   * Gets position of the vector.
   */
  public DVector getPosition() {
    return model.getDVector(0, 1);
  }

  /**
   * Sets position of the vector.
   */
  public void setPosition(DVector position) {
    model.setDVector(0, 1, position);
    update();
  }

  /**
   * Gets the color of the force arrow.
   */
  public Color getArrowColor() {
    return v.getArrowColor();
  }

  /**
   * Sets the color of the force arrow.
   */
  public void setArrowColor(Color arrowColor) {
    v.setArrowColor(arrowColor);
    vx.setArrowColor(arrowColor);
    vy.setArrowColor(arrowColor);
  }

  /**
   * Getter for property arrowScale.
   *
   * @return Value of property arrowScale.
   */
  public double getArrowScale() {
    return v.getArrowScale();
  }

  /**
   * Setter for property arrowScale.
   *
   * @param arrowScale New value of property arrowScale.
   */
  public void setArrowScale(double arrowScale) {
    v.setArrowScale(arrowScale);
    vx.setArrowScale(arrowScale);
    vy.setArrowScale(arrowScale);
  }

  /**
   * gets length of arrow tip relative to arrow length
   */
  public double getArrowTipLength() {
    return v.getArrowTipLength();
  }

  /**
   * sets length of arrow tip relative to arrow length
   */
  public void setArrowTipLength(double arrowTipLength) {
    v.setArrowTipLength(arrowTipLength);
    vx.setArrowTipLength(arrowTipLength);
    vy.setArrowTipLength(arrowTipLength);
  }

  /**
   * gets width of arrow tip relative to arrow length
   */
  public double getArrowTipWidth() {
    return v.getArrowTipWidth();
  }

  /**
   * sets width of arrow tip relative to arrow length
   */
  public void setArrowTipWidth(double arrowTipWidth) {
    v.setArrowTipWidth(arrowTipWidth);
    vx.setArrowTipWidth(arrowTipWidth);
    vy.setArrowTipWidth(arrowTipWidth);
  }

  /**
   * Sets the initial and actual position to pos.
   */
  public void setInitialPosition(DVector pos) {
    setInitialValue(0, pos.x());
    setInitialValue(1, pos.y());
  }

  /**
   * Sets the initial and actual vector to vec.
   */
  public void setInitialVector(DVector vec) {
    setInitialValue(2, vec.x());
    setInitialValue(3, vec.y());
  }

  /**
   * Gets value of showMarker flag.<br>
   * If true, the latest data is displayed with a marker.
   */
  public boolean isShowMarker() {
    return showMarker;
  }

  /**
   * Sets value of showMarker flag.<br>
   * If true, the latest data is displayed with a marker.
   */
  public void setShowMarker(boolean showMarker) {
    this.showMarker = showMarker;
    update();
  }

  /**
   * Gets pixel size of marker.
   */
  public int getMarkerSize() {
    return this.dotSize;
  }

  /**
   * Sets pixel size of marker.
   */
  public void setMarkerSize(int dotSize) {
    this.dotSize = dotSize;
    update();
  }

  /**
   * Gets value of polar flag.<br>
   * If true, the vector is decomposed into radial and tangential parts.
   */
  public boolean isPolar() {
    return polar;
  }

  /**
   * Sets value of polar flag.<br>
   * If true, the vector is decomposed into radial and tangential parts.
   */
  public void setPolar(boolean polar) {
    this.polar = polar;
    update();
  }

  /**
   * Draws point at (x,y) as a filled circle of fixed pixel size.
   */
  protected void drawPoint(Graphics2D g2, double x, double y) {
    double rx = dotSize / 2.0 * trafo.getScalingX();
    double ry = dotSize / 2.0 * trafo.getScalingY();

    // circles of radius 1 sometimes disappear ("smaller than a pixel")
    // in that case I use rectangles
    if (dotSize == 1) {
      Rectangle2D rect = new Rectangle2D.Double(x - rx, y - ry, 2 * rx, 2 * ry);
      g2.fill(rect);
    } else {
      Ellipse2D circ = new Ellipse2D.Double(x - rx, y - ry, 2 * rx, 2 * ry);
      g2.fill(circ);
    }
  }
}
